# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Dict, Literal, Optional, Union

import lightning.pytorch as L
import torch
import torch.distributed
from megatron.core import InferenceParams, parallel_state, tensor_parallel
from megatron.core.fusions.fused_layer_norm import FusedLayerNorm
from megatron.core.models.bert.bert_lm_head import BertLMHead as MCoreBertLMHead
from megatron.core.models.bert.pooler import Pooler
from megatron.core.optimizer import OptimizerConfig
from megatron.core.packed_seq_params import PackedSeqParams
from megatron.core.transformer.spec_utils import ModuleSpec, build_module
from megatron.core.transformer.transformer_block import TransformerBlock
from megatron.core.transformer.transformer_config import TransformerConfig
from megatron.core.transformer.transformer_layer import TransformerLayer, TransformerLayerSubmodules
from megatron.core.transformer.utils import get_linear_layer as mcore_get_linear_layer
from megatron.core.utils import get_batch_on_this_cp_rank, make_viewless_tensor
from torch import Tensor, nn

from nemo.collections.llm import fn
from nemo.collections.llm.bert.loss import BERTLossReduction
from nemo.collections.llm.bert.model.bert_spec import (
    get_bert_layer_local_spec_postln,
    get_bert_layer_with_transformer_engine_spec_postln,
)
from nemo.lightning import get_vocab_size, io
from nemo.lightning.pytorch.optim import MegatronOptimizerModule, OptimizerModule

HAVE_TE = True
try:
    import transformer_engine  # noqa: F401 pylint: disable=W0611
    from megatron.core.models.bert import bert_layer_specs
    from megatron.core.models.bert.bert_model import BertModel as MCoreBert
except (ImportError, ModuleNotFoundError):
    HAVE_TE = False
    MCoreBert = TransformerLayer  # Place holder for import checking. BERT requires TE installed.

if TYPE_CHECKING:
    from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec


def bert_data_step(dataloder_iter) -> Dict[str, torch.Tensor]:
    """Setup BERT dataloader batch."""
    batch = next(dataloder_iter)

    _batch: dict
    if isinstance(batch, tuple) and len(batch) == 3:
        _batch = batch[0]
    else:
        _batch = batch

    required_keys = set()
    required_keys.add("padding_mask")
    if parallel_state.is_pipeline_first_stage():
        required_keys.add("text")
    if parallel_state.is_pipeline_last_stage():
        required_keys.update(("labels", "loss_mask", "types", "is_random"))

    _batch = {key: val.cuda(non_blocking=True) if key in required_keys else None for key, val in _batch.items()}
    # slice batch along sequence dimension for context parallelism
    output = get_batch_on_this_cp_rank(_batch)

    return output


def bert_forward_step(model: L.LightningModule, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
    """
    This subsets the batch keys to the ones actually used by forward pass of the model,
    and then calls the model's forward pass. if "cu_seqsens" are defined in the batch,
    then the packed sequence parameters are also passed to the model for forward pass efficiency.
    """
    forward_args = {
        "input_ids": batch["text"],
        "attention_mask": batch["padding_mask"],
        "lm_labels": batch["labels"],
        "loss_mask": batch["loss_mask"],
    }

    if model.config.num_tokentypes != 0:
        forward_args["tokentype_ids"] = batch["types"]

    if "cu_seqlens" in batch:
        forward_args["packed_seq_params"] = get_packed_seq_params(batch)

    return model(**forward_args)


def default_layer_spec(config: "BertConfig") -> ModuleSpec:
    """
    Return MCore layer spec based on the bert type.
    For bert_type == 'megatron', use mcore's default layer spec;
    For bert_type == 'huggingface', use Post-LayerNorm layer spec.
    """
    bert_type = config.bert_type
    assert (
        bert_type == 'megatron' or bert_type == 'huggingface'
    ), f'Unknown bert type {bert_type}, supported type for bert model is: megatron, huggingface'
    if HAVE_TE:
        if bert_type == 'megatron':
            return bert_layer_specs.bert_layer_with_transformer_engine_spec
        else:
            return get_bert_layer_with_transformer_engine_spec_postln()

    if bert_type == 'megatron':
        return bert_layer_specs.bert_layer_local_spec
    else:
        return get_bert_layer_local_spec_postln()


@dataclass
class BertConfig(TransformerConfig, io.IOMixin):
    """ "Model config for BERT model. Adpated from megatron.core.models.bert.bert_model.BertModel"""

    fp16_lm_cross_entropy: bool = False
    parallel_output: bool = True
    share_embeddings_and_output_weights: bool = True
    position_embedding_type: Literal["learned_absolute", "rope"] = "learned_absolute"
    rotary_base: int = 10000
    rotary_percent: float = 1.0
    seq_len_interpolation_factor: Optional[float] = None
    seq_length: int = 512
    attention_softmax_in_fp32: bool = False
    masked_softmax_fusion: bool = True
    deallocate_pipeline_outputs = True
    make_vocab_size_divisible_by: int = 128

    transformer_layer_spec: Union[ModuleSpec, Callable[["BertConfig"], ModuleSpec]] = default_layer_spec
    forward_step_fn: Callable = bert_forward_step
    data_step_fn: Callable = bert_data_step

    bert_type: Literal["megatron", "huggingface"] = "megatron"
    add_pooler: bool = True
    bert_binary_head: bool = True
    add_lm_head: bool = True
    num_tokentypes: float = None
    mask_vocab_padding_tokens: bool = False

    def configure_model(self, tokenizer, vp_stage: Optional[int] = None) -> "MCoreBertModelWrapperWithPostLNSupport":
        """Configure the BERT Model.
        For bert_type == 'megatron', num_tokentypes in embedding is controlled by whether model has binary head.
        For bert_type == 'huggingface', tokentypes embedding is always added with num_tokentypes = 2.
        """
        vp_size = self.virtual_pipeline_model_parallel_size
        if vp_size:
            p_size = self.pipeline_model_parallel_size
            assert (
                self.num_layers // p_size
            ) % vp_size == 0, "Make sure the number of model chunks is the same across all pipeline stages."
        from megatron.core import parallel_state

        transformer_layer_spec = self.transformer_layer_spec
        if not isinstance(transformer_layer_spec, ModuleSpec):
            transformer_layer_spec = transformer_layer_spec(self)

        if self.num_tokentypes is None:
            self.num_tokentypes = 2 if self.bert_binary_head else 0

        # During fake lightning initialization, pass 0 to bypass the assertion that vp_stage must be
        # non-None when using virtual pipeline model parallelism
        vp_stage = vp_stage or 0
        return MCoreBertModelWrapperWithPostLNSupport(
            bert_type=self.bert_type,
            add_pooler=self.add_pooler,
            config=self,
            num_tokentypes=self.num_tokentypes,
            transformer_layer_spec=transformer_layer_spec,
            vocab_size=get_vocab_size(self, tokenizer.vocab_size, self.make_vocab_size_divisible_by),
            tokenizer=tokenizer if self.mask_vocab_padding_tokens else None,
            max_sequence_length=self.seq_length,
            pre_process=parallel_state.is_pipeline_first_stage(ignore_virtual=False, vp_stage=vp_stage),
            post_process=parallel_state.is_pipeline_last_stage(ignore_virtual=False, vp_stage=vp_stage),
            fp16_lm_cross_entropy=self.fp16_lm_cross_entropy,
            parallel_output=self.parallel_output,
            share_embeddings_and_output_weights=self.share_embeddings_and_output_weights,
            position_embedding_type=self.position_embedding_type,
            rotary_percent=self.rotary_percent,
            # TODO: MCore bert not have rotary base
            seq_len_interpolation_factor=self.seq_len_interpolation_factor,
            add_binary_head=self.bert_binary_head,
            return_embeddings=False,  # TODO
            vp_stage=vp_stage,
        )


class MCoreBertModelWrapperWithPostLNSupport(MCoreBert):
    """
    This class is used for working with HF Bert Checkpoints. These checkpoints
    by default have post layer norm, while the vanilla mcore bert model does not support it.
    when bert_type is set to 'huggingface', it will initialize post layer norm BERT model.
    """

    def __init__(
        self, bert_type='megatron', add_pooler=True, tokenizer: Optional["TokenizerSpec"] = None, *args, **kwargs
    ):
        super(MCoreBertModelWrapperWithPostLNSupport, self).__init__(*args, **kwargs)
        self.add_pooler = add_pooler
        self.bert_type = bert_type
        self.tokenizer = tokenizer

        assert (
            self.bert_type == 'megatron' or self.bert_type == 'huggingface'
        ), f'bert_type should either be megatron or huggingface, but got {self.bert_type}.'

        # Transformer.
        self.encoder = TransformerBlockWithPostLNSupport(
            config=self.config,
            spec=self.transformer_layer_spec,
            pre_process=self.pre_process,
            post_process=self.post_process,
            post_layer_norm=True if self.bert_type == 'megatron' else False,
            bert_type=self.bert_type,
            vp_stage=kwargs.get('vp_stage', None),
        )

        # Megatron Core always initializes position embeddings and tokentype_embeddings in fp32
        # We cast the embedding dtype based on the target dtype in config
        if self.pre_process:
            self.embedding.to(self.config.params_dtype)

        # In Megatron-LM, pooler is added only if add_binary_head=True.
        # We make it independent to support HF variances.
        if self.add_pooler:
            self.pooler = Pooler(
                self.config.hidden_size, self.config.init_method, self.config, self.config.sequence_parallel
            ).to(self.config.params_dtype)

        # Output
        if self.post_process:
            # TODO: Make sure you are passing in the mpu_vocab_size properly
            self.lm_head = None
            if self.config.add_lm_head:
                self.lm_head = MCoreBertLMHead(
                    self.config.hidden_size,
                    self.config,
                )

            self.output_layer = tensor_parallel.ColumnParallelLinear(
                self.config.hidden_size,
                self.vocab_size,
                config=self.config,
                init_method=self.config.init_method,
                bias=True,
                skip_bias_add=True,
                gather_output=not self.parallel_output,
                skip_weight_param_allocation=self.pre_process and self.share_embeddings_and_output_weights,
            )

            self.binary_head = None
            if self.add_binary_head:
                # TODO: Should switch this to TE ?
                self.binary_head = mcore_get_linear_layer(
                    self.config.hidden_size, 2, self.config.init_method, self.config.perform_initialization
                )
        if self.pre_process or self.post_process:
            self.setup_embeddings_and_output_layer()

    def forward(
        self,
        input_ids: Tensor,
        attention_mask: Tensor,
        tokentype_ids: Tensor = None,
        lm_labels: Tensor = None,
        loss_mask: Tensor = None,
        inference_params=None,
        hidden_states_only=False,
    ):
        """Forward function of BERT model

        Forward function of the BERT Model This function passes the input tensors
        through the embedding layer, and then the encoder and finally into the post
        processing layer (optional).

        It either returns the Loss values if labels are given  or the final hidden units
        """
        original_post_process = self.post_process

        # We set this to false since we just want to get the hidden states from the encoder
        self.post_process = False
        hidden_states = super().forward(input_ids, attention_mask, tokentype_ids, lm_labels, inference_params)
        self.post_process = original_post_process

        if not self.post_process or hidden_states_only:
            return hidden_states

        if self.return_embeddings:
            embeddings = torch.transpose(hidden_states, 0, 1)
            masks = torch.sum(attention_mask, dim=1)
            # Collect masked embeddings.
            output = torch.zeros(
                size=(embeddings.shape[0], embeddings.shape[2]),
                dtype=torch.float32,
                device=torch.cuda.current_device(),
            )
            for i, (embedding, mask) in enumerate(zip(embeddings, masks)):
                output[i, :] = torch.mean(embedding[1 : mask - 1], dim=0)
            return output

        # logits and loss
        output_weight = None
        if self.share_embeddings_and_output_weights:
            output_weight = self.shared_embedding_or_output_weight()

        hidden_states_after_lm_head = self.lm_head(hidden_states=hidden_states)
        logits, _ = self.output_layer(hidden_states_after_lm_head, weight=output_weight)

        binary_logits = None
        if self.binary_head is not None and self.add_pooler:
            pooled_output = self.pooler(hidden_states, 0)
            binary_logits = self.binary_head(pooled_output)

        if lm_labels is None:
            # [s b h] => [b s h]0
            return {
                'logits': logits.transpose(0, 1).contiguous(),
                'binary_logits': binary_logits,
                'loss_mask': loss_mask,
            }

        # mask vocab padding tokens from sum term of softmax
        if self.tokenizer:
            from megatron.core.tensor_parallel.utils import VocabUtility

            unpadded_vocab_size = self.tokenizer.vocab_size

            get_vocab_range = VocabUtility.vocab_range_from_per_partition_vocab_size
            padded_vocab_size = logits.size()[-1]
            rank = parallel_state.get_tensor_model_parallel_rank()
            world_size = parallel_state.get_tensor_model_parallel_world_size()
            vocab_start_index, _ = get_vocab_range(padded_vocab_size, rank, world_size)  # gets range on this tp rank

            # mask tokens past unpadded_vocab_size. must be offset by where each tp rank's vocab range starts
            mask_start = max(unpadded_vocab_size - vocab_start_index, 0)
            logits[:, :, mask_start:] = float('-inf')

        loss = self.compute_language_model_loss(lm_labels, logits)

        return {
            'lm_loss': loss,
            'binary_logits': binary_logits,
            'loss_mask': loss_mask,
        }


@dataclass
class TransformerLayerSubmodulesWithPostLNSupport(TransformerLayerSubmodules):
    """Wrapper for TransformerLayerSubmodules with additional post-attention LN and post MLP LN"""

    def __init__(self, post_att_layernorm, post_mlp_layernorm, **kwargs):
        super(TransformerLayerSubmodulesWithPostLNSupport, self).__init__(**kwargs)
        self.post_att_layernorm = post_att_layernorm
        self.post_mlp_layernorm = post_mlp_layernorm


class TransformerLayerWithPostLNSupport(TransformerLayer):
    """
    Adapted from mcore's TransformerLayer with additional post-attention LN and
    post MLP LN support.
    """

    def __init__(self, *args, **kwargs):
        super(TransformerLayerWithPostLNSupport, self).__init__(*args, **kwargs)
        # [Module add: Post attention LN]
        self.post_att_layernorm = build_module(
            self.submodules_config.post_att_layernorm,
            config=self.config,
            hidden_size=self.config.hidden_size,
            eps=self.config.layernorm_epsilon,
        )
        # [Module add: Post MLP LN]
        self.post_mlp_layernorm = build_module(
            self.submodules_config.post_mlp_layernorm,
            config=self.config,
            hidden_size=self.config.hidden_size,
            eps=self.config.layernorm_epsilon,
        )

    def forward(
        self,
        hidden_states,
        attention_mask,
        context=None,
        context_mask=None,
        rotary_pos_emb=None,
        inference_params=None,
        packed_seq_params=None,
        **kwargs,
    ):
        """
        Perform a forward pass through the transformer layer.
        Perform post-attention LN and post MLP LN if module exists.

        This method implements the core computation of a transformer layer, including
        self-attention, cross-attention (if applicable), and feed-forward operations.

        Args:
            hidden_states (Tensor): Input tensor of shape [s, b, h] where s is sequence length,
                b is batch size, and h is hidden size.
            attention_mask (Tensor): Mask tensor for self-attention.
            context (Tensor, optional): Context tensor for cross-attention.
            context_mask (Tensor, optional): Mask tensor for cross-attention.
            rotary_pos_emb (Tensor, optional): Rotary positional embeddings.
            inference_params (object, optional): Parameters for inference-time optimizations.
            packed_seq_params (object, optional): Parameters for packed sequence processing.

        Returns:
            Tuple[Tensor, Tensor]: A tuple containing:
                output (Tensor): Transformed hidden states of shape [s, b, h].
                context (Tensor): Updated context tensor if cross-attention is used,
                otherwise None.
        """
        # hidden_states: [s, b, h]

        # Residual connection.
        residual = hidden_states

        # Optional Input Layer norm
        input_layernorm_output = self.input_layernorm(hidden_states)

        # Self attention.
        attention_output_with_bias = self.self_attention(
            input_layernorm_output,
            attention_mask=attention_mask,
            inference_params=inference_params,
            rotary_pos_emb=rotary_pos_emb,
            packed_seq_params=packed_seq_params,
        )

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.self_attn_bda(self.training, self.config.bias_dropout_fusion)(
                attention_output_with_bias, residual, self.hidden_dropout
            )

        # Residual connection.
        residual = hidden_states

        # Post-LN after Self Attention
        hidden_states = self.post_att_layernorm(hidden_states)

        # Optional Layer norm after self-attention
        pre_cross_attn_layernorm_output = self.pre_cross_attn_layernorm(hidden_states)

        # Cross attention.
        attention_output_with_bias = self.cross_attention(
            pre_cross_attn_layernorm_output,
            attention_mask=context_mask,
            key_value_states=context,
            inference_params=inference_params,
        )

        if isinstance(attention_output_with_bias, dict) and "context" in attention_output_with_bias:
            context = attention_output_with_bias["context"]

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.cross_attn_bda(self.training, self.config.bias_dropout_fusion)(
                attention_output_with_bias, residual, self.hidden_dropout
            )

        # Residual connection.
        residual = hidden_states

        # Optional Layer norm post the cross-attention.
        pre_mlp_layernorm_output = self.pre_mlp_layernorm(hidden_states)

        # MLP.
        mlp_output_with_bias = self.mlp(pre_mlp_layernorm_output)

        # TODO: could we move `bias_dropout_add_exec_handler` itself
        # inside the module provided in the `bias_dropout_add_spec` module?
        with self.bias_dropout_add_exec_handler():
            hidden_states = self.mlp_bda(self.training, self.config.bias_dropout_fusion)(
                mlp_output_with_bias, residual, self.hidden_dropout
            )

        # Post-LN after MLP
        hidden_states = self.post_mlp_layernorm(hidden_states)

        # Jit compiled function creates 'view' tensor. This tensor
        # potentially gets saved in the MPU checkpoint function context,
        # which rejects view tensors. While making a viewless tensor here
        # won't result in memory savings (like the data loader, or
        # p2p_communication), it serves to document the origin of this
        # 'view' tensor.
        output = make_viewless_tensor(inp=hidden_states, requires_grad=hidden_states.requires_grad, keep_graph=True)

        return output, context


class TransformerBlockWithPostLNSupport(TransformerBlock):
    """Adapted from mcore's TransformerBlock with additional post-attention LN and post MLP LN support."""

    def __init__(self, bert_type='megatron', *args, **kwargs):

        super(TransformerBlockWithPostLNSupport, self).__init__(*args, **kwargs)
        self.transformer_block_type = bert_type
        if self.transformer_block_type == 'huggingface':
            # Initial LayerNorm is needed for converting the LN after the HF's Bert Embedding modules:
            # https://github.com/huggingface/transformers/tree/main/src/transformers/models/bert/modeling_bert.py#L170
            # megatron's embedding module does not need the additional LN.
            self.initial_layernorm = FusedLayerNorm(
                config=self.config, hidden_size=self.config.hidden_size, eps=self.config.layernorm_epsilon
            ).to(self.config.params_dtype)

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Tensor,
        context: Tensor = None,
        context_mask: Tensor = None,
        rotary_pos_emb: Tensor = None,
        inference_params: InferenceParams = None,
        packed_seq_params: PackedSeqParams = None,
        **kwargs,
    ):
        """
        Perform the forward pass through the transformer block.
        Perform additional post-attention LN and post MLP LN support if needed.

        This method handles the core computation of the transformer, including
        self-attention, optional cross-attention, and feed-forward operations.

        Args:
            hidden_states (Tensor): Input tensor of shape [s, b, h] where s is the
                sequence length, b is the batch size, and h is the hidden size.
            attention_mask (Tensor): Boolean tensor of shape [1, 1, s, s] for masking
                self-attention.
            context (Tensor, optional): Context tensor for cross-attention.
            context_mask (Tensor, optional): Mask for cross-attention context
            rotary_pos_emb (Tensor, optional): Rotary positional embeddings.
            inference_params (InferenceParams, optional): Parameters for inference-time
                optimizations.
            packed_seq_params (PackedSeqParams, optional): Parameters for packed sequence
                processing.

        Returns:
            Union[Tensor, Tuple[Tensor, Tensor]]: The output hidden states tensor of shape
            [s, b, h], and optionally the updated context tensor if cross-attention is used.
        """
        # hidden_states (float): [s, b, h]
        # attention_mask (bool): [1, 1, s, s]
        if not self.pre_process:
            # See set_input_tensor()
            hidden_states = self.input_tensor
        if self.transformer_block_type == 'huggingface':
            hidden_states = self.initial_layernorm(hidden_states)
        return super(TransformerBlockWithPostLNSupport, self).forward(
            hidden_states, attention_mask, context, context_mask, rotary_pos_emb, inference_params, packed_seq_params
        )


class BertModel(L.LightningModule, io.IOMixin, io.ConnectorMixin, fn.FNMixin):
    """Bert Lightning Module"""

    def __init__(
        self,
        config: BertConfig,
        # TODO: Add transformer_layer_spec when we update mcore
        optim: Optional[OptimizerModule] = None,
        tokenizer: Optional["TokenizerSpec"] = None,
        model_transform: Optional[Callable[[nn.Module], nn.Module]] = None,
    ):
        # Megatron-LM's BERT implementation has high dependency on TE, and it is not possible
        # to instantiate the MCore BERT without TE package.
        # Few issues there: 1. bert_layer_specs.py is not TE dependency-free.
        #                  2. in bert_model.py _sanity_check_attention_and_get_attn_mask_dimension() checks on
        #                     if transformer_layer_spec is identical to bert_layer_local_spec to determine if TE is
        #                     required; since in NeMo we use customized bert layer spec, it will always assume this
        #                     if using TE.
        # We need to address the above two issues to enable TE-Free NeMo BERT.
        assert HAVE_TE, "NeMo BERT requires Transformer Engine to be installed."
        super().__init__()
        self.config = config
        self.tokenizer = tokenizer
        self.optim = optim or MegatronOptimizerModule(config=OptimizerConfig(lr=1e-4, use_distributed_optimizer=True))
        self.optim.connect(self)  # This will bind the `configure_optimizers` method
        self.model_transform = model_transform
        self._training_loss_reduction = None
        self._validation_loss_reduction = None

    def configure_model(self, vp_stage: Optional[int] = None) -> None:
        """Setup the BERT Model based on config definition."""
        if not hasattr(self, "module"):
            self.module = self.config.configure_model(self.tokenizer, vp_stage)

    def forward(
        self,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        """Call the forward method of the underlying model, and return whatever it outputs."""
        output_tensor = self.module(*args, **kwargs)  # for now just pass through to the underlying model
        return output_tensor

    def data_step(self, dataloader_iter) -> Dict[str, torch.Tensor]:  # pylint: disable=C0115,C0116
        return self.config.data_step_fn(dataloader_iter)

    def forward_step(self, batch) -> torch.Tensor:  # pylint: disable=C0115,C0116
        return self.config.forward_step_fn(self, batch)

    def training_step(self, batch, batch_idx=None) -> torch.Tensor:  # pylint: disable=C0115,C0116
        # In mcore the loss-function is part of the forward-pass (when labels are provided)
        return self.forward_step(batch)

    def validation_step(self, batch, batch_idx=None) -> torch.Tensor:  # pylint: disable=C0115,C0116
        # In mcore the loss-function is part of the forward-pass (when labels are provided)
        return self.forward_step(batch)

    @property
    def training_loss_reduction(self) -> BERTLossReduction:  # pylint: disable=C0115,C0116
        if not self._training_loss_reduction:
            self._training_loss_reduction = BERTLossReduction(add_sop_loss=self.config.bert_binary_head)

        return self._training_loss_reduction

    @property
    def validation_loss_reduction(self) -> BERTLossReduction:  # pylint: disable=C0115,C0116
        if not self._validation_loss_reduction:
            self._validation_loss_reduction = BERTLossReduction(
                validation_step=True, add_sop_loss=self.config.bert_binary_head
            )

        return self._validation_loss_reduction


def get_packed_seq_params(batch: Dict[str, torch.Tensor]) -> PackedSeqParams:
    """
    Get the packed sequence parameters for the given batch.
    This function should only be called if `cu_seqlens` is defined in the batch.

    Args:
        batch (dict): The input batch containing the following keys:
            - cu_seqlens (torch.Tensor): The sequence lengths of the input batch.
            - cu_seqlens_argmin (torch.Tensor, optional): The minimum sequence length index.
            - max_seqlen (torch.Tensor, optional): The maximum sequence length.

    Returns:
        PackedSeqParams: The packed sequence parameters containing the following attributes:
            - cu_seqlens_q (torch.Tensor): The sequence lengths for query.
            - cu_seqlens_kv (torch.Tensor): The sequence lengths for key and value.
            - max_seqlen_q (torch.Tensor, optional): The maximum sequence length for query.
            - max_seqlen_kv (torch.Tensor, optional): The maximum sequence length for key and value.
            - qkv_format (str): The format of query, key, and value tensors.

    """
    cu_seqlens = batch["cu_seqlens"].squeeze()  # remove batch size dimension (mbs=1)
    # remove -1 "paddings" added in collate_fn
    if cu_seqlens_argmin := batch.get("cu_seqlens_argmin", None) is not None:
        # pre-compute cu_seqlens_argmin in dataset class for perf
        cu_seqlens = cu_seqlens[: cu_seqlens_argmin.item()]
    else:
        cu_seqlens = cu_seqlens[: torch.argmin(cu_seqlens)]

    # pre-compute max_seqlens in dataset class for perf
    max_seqlen = batch["max_seqlen"].squeeze() if "max_seqlen" in batch else None

    # these args are passed eventually into TEDotProductAttention.forward()
    return PackedSeqParams(
        cu_seqlens_q=cu_seqlens,
        cu_seqlens_kv=cu_seqlens,
        max_seqlen_q=max_seqlen,
        max_seqlen_kv=max_seqlen,
        qkv_format="thd",
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from contextlib import contextmanager
from dataclasses import dataclass, fields
from typing import Generator, Literal, TypeVar, Union

import torch
from lightning.pytorch.plugins.precision import Precision
from torch.nn import Module
from torch.optim import Optimizer

from nemo.utils import logging

AnyT = TypeVar("AnyT")


def get_optim_config(optimizer: Optimizer):
    """Extract optimizer configurations from a Megatron optimizer.

    Args:
        optimizer: A torch.optim.Optimizer instance

    Yields:
        Optimizer configurations
    """
    extract_config = lambda x: x.config
    try:
        from megatron.core.optimizer import ChainedOptimizer

        if isinstance(optimizer.mcore_optimizer, ChainedOptimizer):
            opts = optimizer.mcore_optimizer.chained_optimizers
        else:
            opts = [optimizer.mcore_optimizer]
        yield from map(extract_config, opts)
    except:
        raise ValueError("Failed to extract optimizer config from module.")


@dataclass
class DtypeConfig:
    """Configuration class for mixed precision training settings.

    Contains settings for FP32/FP16/BF16 training, FP8 training.
    """

    fp32: bool = False
    fp16: bool = False
    bf16: bool = False
    params_dtype: torch.dtype = None
    pipeline_dtype: torch.dtype = None
    autocast_dtype: torch.dtype = None
    autocast_enabled: bool = False
    grad_reduce_in_fp32: bool = True
    # fp8 related
    fp8: str = None
    fp8_recipe: str = "delayed"
    # fp4 related
    fp4: str = None
    fp4_recipe: str = "nvfp4"
    first_last_layers_bf16: bool = False
    fp8_margin: int = 0
    fp8_amax_history_len: int = 1
    fp8_amax_compute_algo: str = "most_recent"
    fp8_wgrad: bool = True
    fp8_dot_product_attention: bool = False
    fp8_multi_head_attention: bool = False
    fp8_param: bool = True
    fp8_param_gather: bool = True
    # FP16 Loss scaling
    loss_scale: float = (None,)
    initial_loss_scale: float = (None,)
    min_loss_scale: float = (None,)
    loss_scale_window: float = (None,)
    hysteresis: float = (None,)
    num_layers_at_start_in_bf16: int = 0
    num_layers_at_end_in_bf16: int = 0
    reuse_grad_buf_for_mxfp8_param_ag: bool = False


class MegatronMixedPrecision(Precision):
    """Plugin for mixed precision training with Megatron models.

    Handles conversion of model parameters and inputs/outputs between different precisions,
    and manages mixed precision training settings.
    """

    def __init__(
        self,
        precision: Literal["16-mixed", "bf16-mixed", "32"],
        params_dtype: torch.dtype = None,
        pipeline_dtype: torch.dtype = None,
        autocast_dtype: torch.dtype = None,
        autocast_enabled: bool = False,
        grad_reduce_in_fp32: bool = True,
        # fp8 related,
        fp8: str = None,
        fp8_recipe: str = "delayed",  # "tensorwise", "delayed", "mxfp8" (for Blackwell only)
        first_last_layers_bf16: bool = False,
        fp8_margin: int = 0,
        fp8_amax_history_len: int = 1,
        fp8_amax_compute_algo: str = "most_recent",
        fp8_wgrad: bool = True,
        fp8_dot_product_attention: bool = False,
        fp8_multi_head_attention: bool = False,
        fp8_params: bool = None,
        fp8_param_gather: bool = None,
        # fp4 related
        fp4: str = None,
        fp4_recipe: str = "nvfp4",
        fp16_loss_scale: float = None,
        fp16_initial_loss_scale: float = 4294967296,
        fp16_min_loss_scale: float = 1.0,
        fp16_loss_scale_window: int = 1000,
        fp16_hysteresis: int = 2,
        num_layers_at_start_in_bf16: int = 0,
        num_layers_at_end_in_bf16: int = 0,
        reuse_grad_buf_for_mxfp8_param_ag: bool = False,
    ) -> None:
        if fp8_params is not None:
            logging.warning(
                "fp8_params is deprecated and will be removed in a future release, use fp8_param_gather instead"
            )
            if fp8_param_gather is not None and fp8_param_gather != fp8_params:
                raise ValueError(
                    "Getting conflicting values for fp8_params and fp8_param_gather. Please only set fp8_param_gather."
                )
            fp8_param_gather = fp8_params
        elif fp8_param_gather is None:
            fp8_param_gather = False

        if isinstance(precision, int):
            precision = str(precision)

        dtype = torch.bfloat16 if precision in ['bf16', 'bf16-mixed'] else torch.float32
        self.dtype_config = DtypeConfig(
            fp32=precision in ['fp32', '32'],
            fp16=precision in ['fp16', 'fp16-mixed', '16', '16-mixed'],
            bf16=precision in ['bf16', 'bf16-mixed'],
            params_dtype=params_dtype or torch.float32,
            pipeline_dtype=pipeline_dtype or dtype,
            autocast_dtype=autocast_dtype or dtype,
            autocast_enabled=autocast_enabled,
            grad_reduce_in_fp32=grad_reduce_in_fp32,
            fp8=fp8,
            fp8_recipe=fp8_recipe,
            first_last_layers_bf16=first_last_layers_bf16,
            fp8_margin=fp8_margin,
            fp8_amax_history_len=fp8_amax_history_len,
            fp8_amax_compute_algo=fp8_amax_compute_algo,
            fp8_wgrad=fp8_wgrad,
            fp8_dot_product_attention=fp8_dot_product_attention,
            fp8_multi_head_attention=fp8_multi_head_attention,
            fp8_param=fp8_param_gather,
            fp8_param_gather=fp8_param_gather,
            fp4=fp4,
            fp4_recipe=fp4_recipe,
            num_layers_at_start_in_bf16=num_layers_at_start_in_bf16,
            num_layers_at_end_in_bf16=num_layers_at_end_in_bf16,
            reuse_grad_buf_for_mxfp8_param_ag=reuse_grad_buf_for_mxfp8_param_ag,
            # fp16 loss scale
            loss_scale=fp16_loss_scale,
            initial_loss_scale=fp16_initial_loss_scale,
            min_loss_scale=fp16_min_loss_scale,
            loss_scale_window=fp16_loss_scale_window,
            hysteresis=fp16_hysteresis,
        )
        super().__init__()
        if self.dtype_config.fp16:
            self.precision = "16-mixed"
        elif self.dtype_config.bf16:
            self.precision = "bf16-mixed"
        else:
            self.precision = "32-true"

    def convert_module(self, module: Module) -> Module:
        """Convert the module parameters to the precision type this plugin handles.

        This is optional and depends on the precision limitations during optimization.

        """
        from megatron.core.transformer.module import Float16Module
        from megatron.core.utils import get_model_config

        if self.dtype_config.fp16 or self.dtype_config.bf16:
            # Patch config options
            config = get_model_config(module.module)
            config.fp16 = self.dtype_config.fp16
            config.bf16 = self.dtype_config.bf16
            # Avoid rewrapping the module if it's already of type Float16Module
            if hasattr(module, "module"):
                if not isinstance(module.module, Float16Module):
                    module.module = Float16Module(config, module.module)
            elif not isinstance(module, Float16Module):
                module = Float16Module(config, module)

        return module

    def convert_optimizer(self, optimizer: Optimizer) -> Optimizer:
        """Convert the optimizer parameters to the precision type this plugin handles.

        This is optional and depends on the precision limitations during optimization.

        """
        for optim_config in get_optim_config(optimizer):
            assert optim_config.bf16 == self.dtype_config.bf16, "BF16 model/optim config mismatch"
            assert optim_config.fp16 == self.dtype_config.fp16, "FP16 model/optim config mismatch"
        return optimizer

    def convert_input(self, data: AnyT) -> AnyT:
        """Convert model inputs (forward) to the floating point precision type of this plugin.

        Note: MegatronStrategy will take care of only doing this when:
            parallel_state.is_pipeline_first_stage()

        """
        return data

    def convert_output(self, data: AnyT) -> AnyT:
        """Convert outputs to the floating point precision type expected after model's forward.

        Note: MegatronStrategy will take care of only doing this when:
            parallel_state.is_pipeline_last_stage()

        """
        return data

    @contextmanager
    def forward_context(self) -> Generator[None, None, None]:
        """No explicit precision casting. Inputs are supposed to be manually casted."""
        try:
            yield
        finally:
            pass

    def clip_gradients(
        self,
        optimizer: Optimizer,
        clip_val: Union[int, float] = 0.0,
        gradient_clip_algorithm=None,
    ) -> None:
        """Clip gradients. Raises error if clip_val > 0, otherwise it is a no-op.

        Args:
            optimizer: The optimizer to clip gradients for
            clip_val: The value to clip gradients to
            gradient_clip_algorithm: The algorithm to use for clipping

        Raises:
            ValueError: If clip_val > 0 since gradient clipping is handled by Mcore's optimizer
        """
        if clip_val > 0.0:
            raise ValueError(
                "Gradient clipping is handled in Mcore's optimizer. Use the clip_grad attribute in OptimizerConfig."
            )

    def clip_grad_by_value(self, optimizer: Optimizer, clip_val: Union[int, float]) -> None:
        """Clip gradients by value - it is a no-op.

        Args:
            optimizer: The optimizer to clip gradients for
            clip_val: The value to clip gradients to
        """
        return

    def clip_grad_by_norm(self, optimizer: Optimizer, clip_val: Union[int, float]) -> None:
        """Clip gradients by norm - it is a no-op.

        Args:
            optimizer: The optimizer to clip gradients for
            clip_val: The value to clip gradients to
        """
        return


def update_config_with_dtype_overrides(dtype_config, config):
    """Update a config object with dtype settings from dtype_config.

    Args:
        dtype_config: Source of dtype settings
        config: Config object to update

    Returns:
        Updated config object
    """
    if hasattr(config, "__io__"):
        config.__io__ = update_config_with_dtype_overrides(dtype_config, config.__io__)
    for field in fields(dtype_config):
        if not hasattr(config, field.name):
            continue
        # If we overwrote a value, log a debug message.
        old_val = getattr(config, field.name)
        new_val = getattr(dtype_config, field.name)
        if old_val != new_val:
            setattr(config, field.name, new_val)
            logging.debug(f"Overwrote {type(config).__name__}.{field.name}  {old_val} -> {new_val}")
    return config


__all__ = ["MegatronMixedPrecision"]



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from nemo.collections.asr.parts.k2.utils import make_non_pad_mask


class GradExpNormalize(torch.autograd.Function):
    """Function for fast gradient normalization.
    Typical use case is normalization for mle loss.
    """

    @staticmethod
    def forward(
        ctx, log_probs: torch.Tensor, input_lengths: torch.Tensor, reduction: str = "mean",
    ):
        mask = make_non_pad_mask(input_lengths, log_probs.shape[1])
        probs = log_probs.exp()
        norm_probs = torch.zeros_like(log_probs)
        norm_probs[mask] += probs[mask]
        if reduction == "mean":
            norm_probs /= norm_probs.shape[0]
        ctx.save_for_backward(norm_probs)
        return log_probs

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        return grad_output - grad_output.sum(-1).unsqueeze(-1) * ctx.saved_tensors[0], None, None


class GradInsert(torch.autograd.Function):
    """Function to attach a pre-computed gradient to a tensor.
    Typical use case is gradient computation before calling loss.backward().
    """

    @staticmethod
    def forward(
        ctx, input_tensor: torch.Tensor, output_tensor: torch.Tensor, grad: torch.Tensor, mask: torch.Tensor,
    ):
        assert input_tensor.requires_grad
        assert not output_tensor.requires_grad and not grad.requires_grad

        ctx.save_for_backward(grad, mask)
        return output_tensor

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        saved_grad, mask = ctx.saved_tensors
        # TODO (alaptev): make it work for grad_output with arbitrary shape
        padded_grad_output = torch.zeros(saved_grad.shape[0], dtype=grad_output.dtype, device=grad_output.device)
        padded_grad_output[mask] = grad_output
        return (padded_grad_output * saved_grad.T).T, None, None, None


class PartialGrad(torch.nn.Module):
    """Module for partial gradient computation.
    Useful when computing loss on batch splits to save memory.
    """

    def __init__(self, func: torch.nn.Module):
        super().__init__()
        self.func = func

    def forward(
        self,
        input_tensor: torch.Tensor,
        targets: torch.Tensor,
        input_lengths: torch.Tensor,
        target_lengths: torch.Tensor,
    ):
        # break the gradient chain
        loc_tensor = input_tensor.detach()
        loc_tensor.requires_grad_(True)

        new_tensor, mask = self.func(loc_tensor, targets, input_lengths, target_lengths)
        loc_new_tensor = new_tensor.detach()

        new_tensor.sum().backward()
        grad = loc_tensor.grad

        return GradInsert.apply(input_tensor, loc_new_tensor, grad, mask), mask



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any

from torch import Tensor

from nemo.collections.asr.inference.model_wrappers.asr_inference_wrapper import ASRInferenceWrapper


class CacheAwareASRInferenceWrapper(ASRInferenceWrapper):
    """
    Base class for Cache-Aware inference wrappers.
    It provides a common interface for Cache-Aware models.
    Derived classes MUST implement the following methods:
        - stream_step: Executes a single streaming step.
    """

    def get_input_features(self) -> int:
        """
        Returns the number of channels in the input features.
        Returns:
            (int) number of channels in the input features.
        """
        return self.asr_model.encoder._feat_in

    def get_sampling_frames(self) -> list[int] | int | None:
        """
        It is used for checking to make sure the audio chunk has enough frames to produce at least one output after downsampling.
        Returns:
            (list[int] | int | None) sampling frames for the encoder.
        """
        self.sampling_frames = None
        if hasattr(self.asr_model.encoder, "pre_encode") and hasattr(
            self.asr_model.encoder.pre_encode, "get_sampling_frames"
        ):
            self.sampling_frames = self.asr_model.encoder.pre_encode.get_sampling_frames()
        return self.sampling_frames

    def get_initial_cache_state(self, batch_size: int) -> tuple[Tensor, Tensor, Tensor]:
        """
        Returns the initial cache state for the encoder.
        Returns:
            (tuple[Tensor, Tensor, Tensor]) the initial cache state of the encoder.
        """
        return self.asr_model.encoder.get_initial_cache_state(batch_size=batch_size)

    def get_drop_extra_pre_encoded(self) -> int:
        """
        Returns the number of extra pre-encoded frames to drop.
        Returns:
            (int) drop_extra_pre_encoded.
        """
        return self.asr_model.encoder.streaming_cfg.drop_extra_pre_encoded

    def get_chunk_size(self) -> list[int] | int:
        """
        Returns the chunk size for the encoder.
        Returns:
            (list[int] | int) the chunk size.
        """
        return self.asr_model.encoder.streaming_cfg.chunk_size

    def get_shift_size(self) -> list[int] | int:
        """
        Returns the shift size for the encoder.
        Returns:
            (list[int] | int) the shift size.
        """
        return self.asr_model.encoder.streaming_cfg.shift_size

    def get_pre_encode_cache_size(self) -> list[int] | int:
        """
        Returns the pre-encode cache size for the encoder.
        Returns:
            (list[int] | int) the pre_encode cache size.
        """
        return self.asr_model.encoder.streaming_cfg.pre_encode_cache_size

    def get_subsampling_factor(self) -> int:
        """
        Returns the subsampling factor for the ASR encoder.
        Returns:
            (int) subsampling factor for the ASR encoder model.
        """
        return self.asr_model.encoder.subsampling_factor

    def get_att_context_size(self) -> list:
        """
        Returns the attention context size for the encoder.
        Returns:
            (list) copy of the attention context size.
        """
        return self.asr_model.encoder.att_context_size.copy()

    def set_default_att_context_size(self, att_context_size: list) -> None:
        """
        Set the default attention context size for the encoder.
        The list of the supported look-ahead: [[70, 13], [70, 6], [70, 1], [70, 0]]
        Args:
            att_context_size: (list) the attention context size.
        """
        if hasattr(self.asr_model.encoder, "set_default_att_context_size"):
            self.asr_model.encoder.set_default_att_context_size(att_context_size=att_context_size)
        else:
            raise ValueError("Model does not support multiple lookaheads.")

    def setup_streaming_params(self, chunk_size: int, shift_size: int) -> None:
        """
        Setup the streaming parameters (chunk_size, shift_size) for the encoder.
        Args:
            chunk_size: (int) the chunk size.
            shift_size: (int) the shift size.
        """
        self.asr_model.encoder.setup_streaming_params(chunk_size=chunk_size, shift_size=shift_size)

    def stream_step(self, *args, **kwargs) -> Any:
        """
        Executes a single streaming step.
        Each derived class must implement this method, with arguments and return types specific to that class.
        """
        raise NotImplementedError(
            "`stream_step` method is not implemented. It is required for cache-aware transcribers."
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest

from nemo.collections import llm
from nemo.collections.llm import Llama32EmbeddingConfig1B, LlamaEmbeddingModel
from nemo.collections.llm.api import finetune
from nemo.collections.llm.peft.lora import LoRA
from nemo.collections.llm.recipes import llama_embedding_1b
from nemo.lightning import Trainer
from nemo.lightning.pytorch.callbacks.garbage_collection import GarbageCollectionCallback
from nemo.utils.exp_manager import TimingCallback


class TestLlamaEmbedding_1B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        return llama_embedding_1b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        assert isinstance(model_config, run.Config)
        assert model_config.__fn_or_cls__ == LlamaEmbeddingModel
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == Llama32EmbeddingConfig1B

    def test_trainer(self, recipe_module):
        trainer_config = recipe_module.trainer()
        assert isinstance(trainer_config, run.Config)
        assert trainer_config.__fn_or_cls__ == Trainer
        assert trainer_config.accelerator == "gpu"
        assert trainer_config.devices == 8
        assert trainer_config.num_nodes == 1
        assert trainer_config.max_steps == 1168251
        assert trainer_config.limit_val_batches == 32
        assert trainer_config.limit_test_batches == 50
        assert trainer_config.log_every_n_steps == 10
        assert trainer_config.val_check_interval == 2000
        assert trainer_config.use_distributed_sampler is False

        # Check strategy configuration
        assert isinstance(trainer_config.strategy, run.Config)
        assert trainer_config.strategy.__fn_or_cls__.__name__ == "MegatronStrategy"
        assert trainer_config.strategy.tensor_model_parallel_size == 1
        assert trainer_config.strategy.pipeline_model_parallel_size == 1
        assert trainer_config.strategy.pipeline_dtype is None
        assert trainer_config.strategy.virtual_pipeline_model_parallel_size is None
        assert trainer_config.strategy.context_parallel_size == 2
        assert trainer_config.strategy.sequence_parallel is False

        # Check DDP configuration
        assert trainer_config.strategy.ddp.check_for_nan_in_grad is True
        assert trainer_config.strategy.ddp.grad_reduce_in_fp32 is True
        assert trainer_config.strategy.ddp.overlap_grad_reduce is True
        assert trainer_config.strategy.ddp.overlap_param_gather is True
        assert trainer_config.strategy.ddp.average_in_collective is True

    def test_finetune_recipe(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == LlamaEmbeddingModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == llm.SpecterDataModule  # Different from other Llama models
        assert recipe.data.seq_length == 512  # Default for embedding model
        assert recipe.data.micro_batch_size == 4  # Default for embedding model
        assert recipe.data.global_batch_size == 64  # Default for embedding model
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == LoRA
        assert recipe.peft.dim == 8
        assert recipe.peft.alpha == 16
        assert recipe.optim.config.lr == 1e-4

    def test_finetune_recipe_with_packed_sequence(self, recipe_module):
        with pytest.raises(AssertionError, match='pack_sequence is not supported for Embedding model finetuning.'):
            recipe_module.finetune_recipe(packed_sequence=True)

    def test_finetune_recipe_without_peft(self, recipe_module):
        recipe = recipe_module.finetune_recipe(peft_scheme=None)
        assert recipe.trainer.strategy.tensor_model_parallel_size == 1
        assert recipe.optim.config.lr == 5e-6
        assert not hasattr(recipe, 'peft') or recipe.peft is None

    def test_finetune_recipe_with_invalid_peft(self, recipe_module):
        with pytest.raises(ValueError, match="Unrecognized peft scheme: invalid_scheme"):
            recipe_module.finetune_recipe(peft_scheme="invalid_scheme")

    def test_finetune_performance_optimizations(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        recipe = recipe_module.finetune_performance_optimizations(recipe, peft_scheme='lora')
        assert recipe.trainer.strategy.tensor_model_parallel_size == 1
        assert recipe.peft.target_modules == ['linear_qkv']
        assert any(
            isinstance(cb, run.Config) and cb.__fn_or_cls__ == TimingCallback for cb in recipe.trainer.callbacks
        )
        assert any(
            isinstance(cb, run.Config) and cb.__fn_or_cls__ == GarbageCollectionCallback
            for cb in recipe.trainer.callbacks
        )

    def test_finetune_performance_optimizations_without_peft(self, recipe_module):
        recipe = recipe_module.finetune_recipe(peft_scheme=None)
        recipe = recipe_module.finetune_performance_optimizations(recipe, peft_scheme=None)
        assert recipe.trainer.plugins.grad_reduce_in_fp32 is False
        assert recipe.trainer.strategy.ddp.grad_reduce_in_fp32 is False
        assert recipe.trainer.strategy.ddp.overlap_grad_reduce is True
        assert recipe.trainer.strategy.ddp.overlap_param_gather is True
        assert recipe.trainer.strategy.ddp.average_in_collective is True
        assert any(
            isinstance(cb, run.Config) and cb.__fn_or_cls__.__name__ == "MegatronCommOverlapCallback"
            for cb in recipe.trainer.callbacks
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from typing import Dict, Optional, Tuple

import torch.utils.data
from lhotse.dataset import AudioSamples
from lhotse.dataset.collation import collate_vectors

from nemo.collections.asr.data.audio_to_text_lhotse import TokenizerWrapper
from nemo.collections.asr.parts.utils.asr_multispeaker_utils import speaker_to_target
from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.core.neural_types import AudioSignal, LabelsType, LengthsType, NeuralType


class LhotseSpeechToTextSpkBpeDataset(torch.utils.data.Dataset):
    """
    This dataset is based on BPE datasets from audio_to_text.py. It has the same functionality of LhotseSpeechToTextBpeDataset but also yield speaker target tensor.
    Unlike native NeMo datasets, Lhotse dataset defines only the mapping from
    a CutSet (meta-data) to a mini-batch with PyTorch tensors.
    Specifically, it performs tokenization, I/O, augmentation, and feature extraction (if any).
    Managing data, sampling, de-duplication across workers/nodes etc. is all handled
    by Lhotse samplers instead.
    """

    @property
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        return {
            'audio_signal': NeuralType(('B', 'T'), AudioSignal()),
            'a_sig_length': NeuralType(tuple('B'), LengthsType()),
            'transcripts': NeuralType(('B', 'T'), LabelsType()),
            'transcript_length': NeuralType(tuple('B'), LengthsType()),
            'spk_targets': NeuralType(('B', 'T'), LabelsType()),
            'bg_spk_targets': NeuralType(('B', 'T'), LabelsType()),
        }

    def __init__(self, cfg, tokenizer: TokenizerSpec):
        super().__init__()
        self.tokenizer = TokenizerWrapper(tokenizer)
        self.load_audio = AudioSamples(fault_tolerant=True)
        self.cfg = cfg
        self.num_speakers = self.cfg.get('num_speakers', 4)
        self.num_sample_per_mel_frame = self.cfg.get('num_sample_per_mel_frame', 160)
        self.num_mel_frame_per_asr_frame = self.cfg.get('num_mel_frame_per_asr_frame', 8)
        self.fixed_spk_id = self.cfg.get('fixed_spk_id', None)
        self.inference_mode = self.cfg.get('inference_mode', False)

    def __getitem__(self, cuts) -> Tuple[torch.Tensor, ...]:

        audio, audio_lens, cuts = self.load_audio(cuts)

        tokens = []
        spk_targets = []
        bg_spk_targets = []

        if self.inference_mode:
            return audio, audio_lens, None, None, None, None

        for idx, cut in enumerate(cuts):

            speaker_targets, texts = speaker_to_target(
                a_cut=cut,
                num_speakers=self.num_speakers,
                num_sample_per_mel_frame=self.num_sample_per_mel_frame,
                num_mel_frame_per_asr_frame=self.num_mel_frame_per_asr_frame,
                return_text=True,
            )
            speaker_targets = speaker_targets.transpose(0, 1)[: len(texts)]

            target_speaker_id = random.choice(range(len(texts)))
            non_target_speaker_ids = [i for i in range(len(texts)) if i != target_speaker_id]
            text = texts[target_speaker_id]
            speaker_target = speaker_targets[target_speaker_id]
            bg_speaker_target = speaker_targets[non_target_speaker_ids].sum(dim=0) > 0

            tokens.append(torch.as_tensor(self.tokenizer(text, cut.supervisions[0].language)))
            spk_targets.append(speaker_target)
            bg_spk_targets.append(bg_speaker_target)

        token_lens = torch.tensor([t.size(0) for t in tokens], dtype=torch.long)
        tokens = collate_vectors(tokens, padding_value=0)
        spk_targets = collate_vectors(spk_targets, padding_value=0)
        bg_spk_targets = collate_vectors(bg_spk_targets, padding_value=0)

        return audio, audio_lens, tokens, token_lens, spk_targets, bg_spk_targets



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
from dataclasses import dataclass
from typing import Any, Dict, List

import pytest
import torch
from omegaconf import open_dict
from torch.utils.data import DataLoader, Dataset

from nemo.collections.asr.data.audio_to_text import _speech_collate_fn
from nemo.collections.asr.models.aed_multitask_models import MultiTaskTranscriptionConfig
from nemo.collections.asr.parts.mixins import TranscribeConfig, TranscriptionMixin
from nemo.collections.asr.parts.mixins.transcription import GenericTranscriptionType
from nemo.collections.asr.parts.submodules.multitask_decoding import MultiTaskDecodingConfig
from nemo.collections.asr.parts.utils import Hypothesis


class DummyModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torch.nn.Linear(1, 1)

        self.execution_count = 0
        self.flag_begin = False
        self.flag_end = False

    def forward(self, x):
        # Input: [1, 1] Output = [1, 1
        out = self.encoder(x)
        return out


class DummyDatasetAudioOnly(Dataset):
    def __init__(self, audio_files: List[str], config: Dict):
        self.audio_files = audio_files
        self.config = config

    def __getitem__(self, index):
        data = self.audio_files[index]
        data = torch.tensor([float(data)]).view(1)
        return data

    def __len__(self):
        return len(self.audio_files)


class DummyDataset(Dataset):
    def __init__(self, audio_tensors: List[str], config: Dict = None):
        self.audio_tensors = audio_tensors
        self.config = config

    def __getitem__(self, index):
        data = self.audio_tensors[index]
        samples = torch.tensor(data)
        # Calculate seq length
        seq_len = torch.tensor(samples.shape[0], dtype=torch.long)

        # Dummy text tokens
        text_tokens = torch.tensor([0], dtype=torch.long)
        text_tokens_len = torch.tensor(1, dtype=torch.long)

        return (samples, seq_len, text_tokens, text_tokens_len)

    def __len__(self):
        return len(self.audio_tensors)


@pytest.fixture()
def audio_files(test_data_dir):
    """
    Returns a list of audio files for testing.
    """
    import soundfile as sf

    audio_file1 = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")
    audio_file2 = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an104-mrcb-b.wav")

    audio1, _ = sf.read(audio_file1, dtype='float32')
    audio2, _ = sf.read(audio_file2, dtype='float32')

    return audio1, audio2


class TranscribableDummy(DummyModel, TranscriptionMixin):
    def _transcribe_on_begin(self, audio, trcfg: TranscribeConfig):
        super()._transcribe_on_begin(audio, trcfg)
        self.flag_begin = True

    def _transcribe_input_manifest_processing(self, audio_files: List[str], temp_dir: str, trcfg: TranscribeConfig):
        # Create a dummy manifest
        manifest_path = os.path.join(temp_dir, 'dummy_manifest.json')
        with open(manifest_path, 'w', encoding='utf-8') as fp:
            for audio_file in audio_files:
                entry = {'audio_filepath': audio_file, 'duration': 100000, 'text': ''}
                fp.write(json.dumps(entry) + '\n')

        ds_config = {
            'paths2audio_files': audio_files,
            'batch_size': trcfg.batch_size,
            'temp_dir': temp_dir,
            'num_workers': trcfg.num_workers,
            'channel_selector': trcfg.channel_selector,
        }

        return ds_config

    def _setup_transcribe_dataloader(self, config: Dict) -> DataLoader:
        dataset = DummyDatasetAudioOnly(config['paths2audio_files'], config)

        return DataLoader(
            dataset=dataset,
            batch_size=config['batch_size'],
            num_workers=config['num_workers'],
            pin_memory=False,
            drop_last=False,
        )

    def _transcribe_forward(self, batch: Any, trcfg: TranscribeConfig):
        output = self(batch)
        return output

    def _transcribe_output_processing(self, outputs, trcfg: TranscribeConfig) -> GenericTranscriptionType:
        self.execution_count += 1

        result = []
        for output in outputs:
            result.append(float(output.item()))

        if hasattr(trcfg, 'output_type') and trcfg.output_type == 'dict':
            results = {'output': result}
            return results

        if hasattr(trcfg, 'output_type') and trcfg.output_type == 'dict2':
            results = [{'output': res} for res in result]
            return results

        if hasattr(trcfg, 'output_type') and trcfg.output_type == 'tuple':
            result = tuple(result)
            return result

        # Pass list of results by default
        return result

    def _transcribe_on_end(self, trcfg: TranscribeConfig):
        super()._transcribe_on_end(trcfg)
        self.flag_end = True


@pytest.fixture()
def dummy_model():
    return TranscribableDummy()


class TestTranscriptionMixin:
    @pytest.mark.unit
    def test_constructor_non_instance(self):
        model = DummyModel()
        assert not isinstance(model, TranscriptionMixin)
        assert not hasattr(model, 'transcribe')

    @pytest.mark.unit
    def test_transcribe(self, dummy_model):
        dummy_model = dummy_model.eval()
        dummy_model.encoder.weight.data.fill_(1.0)
        dummy_model.encoder.bias.data.fill_(0.0)

        audio = ['1.0', '2.0', '3.0']
        outputs = dummy_model.transcribe(audio, batch_size=1)
        assert len(outputs) == 3
        assert outputs[0] == 1.0
        assert outputs[1] == 2.0
        assert outputs[2] == 3.0

    @pytest.mark.unit
    def test_transcribe_generator(self, dummy_model):
        dummy_model = dummy_model.eval()
        dummy_model.encoder.weight.data.fill_(1.0)
        dummy_model.encoder.bias.data.fill_(0.0)

        audio = ['1.0', '2.0', '3.0']

        transribe_config = TranscribeConfig(batch_size=1)
        generator = dummy_model.transcribe_generator(audio, override_config=transribe_config)

        outputs = []
        index = 1
        for result in generator:
            outputs.extend(result)
            assert len(result) == 1
            assert len(outputs) == index
            index += 1

        assert len(outputs) == 3
        assert outputs[0] == 1.0
        assert outputs[1] == 2.0
        assert outputs[2] == 3.0

    @pytest.mark.unit
    def test_transcribe_generator_explicit_stop_check(self, dummy_model):
        dummy_model = dummy_model.eval()
        dummy_model.encoder.weight.data.fill_(1.0)
        dummy_model.encoder.bias.data.fill_(0.0)

        audio = ['1.0', '2.0', '3.0']

        transribe_config = TranscribeConfig(batch_size=1)
        generator = dummy_model.transcribe_generator(audio, override_config=transribe_config)

        outputs = []
        index = 1
        while True:
            try:
                result = next(generator)
            except StopIteration:
                break
            outputs.extend(result)
            assert len(result) == 1
            assert len(outputs) == index
            index += 1

        assert len(outputs) == 3
        assert outputs[0] == 1.0
        assert outputs[1] == 2.0
        assert outputs[2] == 3.0

    @pytest.mark.unit
    def test_transcribe_check_flags(self, dummy_model):
        dummy_model = dummy_model.eval()

        audio = ['1.0', '2.0', '3.0']
        dummy_model.transcribe(audio, batch_size=1)
        assert dummy_model.flag_begin
        assert dummy_model.flag_end

    @pytest.mark.unit
    def test_transribe_override_config_incorrect(self, dummy_model):
        # Not subclassing TranscribeConfig
        @dataclass
        class OverrideConfig:
            batch_size: int = 1
            output_type: str = 'dict'

        dummy_model = dummy_model.eval()

        audio = [1.0, 2.0, 3.0]
        override_cfg = OverrideConfig(batch_size=1, output_type='dict')
        with pytest.raises(ValueError):
            _ = dummy_model.transcribe(audio, override_config=override_cfg)

    @pytest.mark.unit
    def test_transribe_override_config_correct(self, dummy_model):
        @dataclass
        class OverrideConfig(TranscribeConfig):
            output_type: str = 'dict'
            verbose: bool = False

        dummy_model = dummy_model.eval()
        dummy_model.encoder.weight.data.fill_(1.0)
        dummy_model.encoder.bias.data.fill_(0.0)

        audio = ['1.0', '2.0', '3.0']
        override_cfg = OverrideConfig(batch_size=1, output_type='dict')
        outputs = dummy_model.transcribe(audio, override_config=override_cfg)

        assert isinstance(outputs, dict)
        assert len(outputs) == 1
        assert dummy_model.execution_count == 3
        assert outputs['output'][0] == 1.0
        assert outputs['output'][1] == 2.0
        assert outputs['output'][2] == 3.0

        # Reset execution count
        dummy_model.execution_count = 0

        override_cfg = OverrideConfig(batch_size=1, output_type='dict2')
        outputs = dummy_model.transcribe(audio, override_config=override_cfg)

        # Output now is list of dict of value each
        assert isinstance(outputs, list)
        assert len(outputs) == 3
        assert dummy_model.execution_count == 3
        assert outputs[0]['output'] == 1.0
        assert outputs[1]['output'] == 2.0
        assert outputs[2]['output'] == 3.0

        # Reset execution count
        dummy_model.execution_count = 0

        # Test tuple
        override_cfg = OverrideConfig(batch_size=1, output_type='tuple')
        outputs = dummy_model.transcribe(audio, override_config=override_cfg)

        assert isinstance(outputs, tuple)
        assert len(outputs) == 1
        assert dummy_model.execution_count == 3
        assert outputs[0][0] == 1.0
        assert outputs[0][1] == 2.0
        assert outputs[0][2] == 3.0

    pytest.mark.with_downloads()

    @pytest.mark.unit
    def test_transcribe_return_hypothesis(self, test_data_dir, fast_conformer_ctc_model):
        audio_file = os.path.join(test_data_dir, "asr", "train", "an4", "wav", "an46-mmap-b.wav")

        # Audio file test
        outputs = fast_conformer_ctc_model.transcribe(audio_file, batch_size=1, return_hypotheses=True)
        assert len(outputs) == 1
        assert isinstance(outputs[0], Hypothesis)

        hyp = outputs[0]
        assert isinstance(hyp.text, str)
        assert isinstance(hyp.y_sequence, torch.Tensor)
        assert isinstance(hyp.alignments, torch.Tensor)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_transcribe_tensor(self, audio_files, fast_conformer_ctc_model):

        audio, _ = audio_files
        # Numpy array test
        outputs = fast_conformer_ctc_model.transcribe(audio, batch_size=1)
        assert len(outputs) == 1
        assert isinstance(outputs[0], Hypothesis)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_transcribe_multiple_tensor(self, audio_files, fast_conformer_ctc_model):

        audio, audio_2 = audio_files
        # Mix second audio to torch.tensor()
        audio_2 = torch.tensor(audio_2)

        # Numpy array test
        outputs = fast_conformer_ctc_model.transcribe([audio, audio_2], batch_size=2)
        assert len(outputs) == 2
        assert isinstance(outputs[0], Hypothesis)
        assert isinstance(outputs[1], Hypothesis)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_transcribe_dataloader(self, audio_files, fast_conformer_ctc_model):

        audio, audio2 = audio_files

        dataset = DummyDataset([audio, audio2])
        collate_fn = lambda x: _speech_collate_fn(x, pad_id=0)
        dataloader = DataLoader(dataset, batch_size=2, shuffle=False, num_workers=0, collate_fn=collate_fn)

        # DataLoader test
        outputs = fast_conformer_ctc_model.transcribe(dataloader, batch_size=1)
        assert len(outputs) == 2
        assert isinstance(outputs[0], Hypothesis)
        assert isinstance(outputs[1], Hypothesis)

    @pytest.mark.unit
    def test_transcribe_return_nbest_rnnt(self, audio_files, fast_conformer_transducer_model):
        fast_conformer_transducer_model.eval()
        audio1, audio2 = audio_files

        orig_decoding_config = copy.deepcopy(fast_conformer_transducer_model.cfg.decoding)

        decoding_config = copy.deepcopy(fast_conformer_transducer_model.cfg.decoding)
        with open_dict(decoding_config):
            decoding_config["strategy"] = "malsd_batch"
            decoding_config["beam"]["beam_size"] = 4
            decoding_config["beam"]["return_best_hypothesis"] = False
            decoding_config["beam"]["allow_cuda_graphs"] = False
        fast_conformer_transducer_model.change_decoding_strategy(decoding_config)

        outputs = fast_conformer_transducer_model.transcribe([audio1, audio2], batch_size=1, timestamps=False)

        assert len(outputs) == 2
        assert all(len(output) >= 1 for output in outputs)
        assert all(isinstance(output, list) for output in outputs)
        assert all(isinstance(hyp, Hypothesis) for output in outputs for hyp in output)

        # Reset the decoding strategy to original
        fast_conformer_transducer_model.change_decoding_strategy(orig_decoding_config)

    @pytest.mark.unit
    def test_transcribe_return_nbest_canary(self, audio_files, canary_1b_flash):
        canary_1b_flash.eval()
        audio1, audio2 = audio_files

        orig_decoding_config = copy.deepcopy(canary_1b_flash.cfg.decoding)

        decoding_config = copy.deepcopy(canary_1b_flash.cfg.decoding)
        with open_dict(decoding_config):
            decoding_config["beam"]["beam_size"] = 4
            decoding_config["beam"]["return_best_hypothesis"] = False
        canary_1b_flash.change_decoding_strategy(decoding_config)

        outputs = canary_1b_flash.transcribe([audio1, audio2], batch_size=1, timestamps=False)

        assert len(outputs) == 2
        assert all(len(output) >= 1 for output in outputs)
        assert all(isinstance(output, list) for output in outputs)
        assert all(isinstance(hyp, Hypothesis) for output in outputs for hyp in output)

        # Reset the decoding strategy to original
        canary_1b_flash.change_decoding_strategy(orig_decoding_config)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_timestamps_with_transcribe(self, audio_files, fast_conformer_ctc_model):
        audio1, audio2 = audio_files

        output = fast_conformer_ctc_model.transcribe([audio1, audio2], timestamps=True)

        # check len of output
        assert len(output) == 2

        # check hypothesis object
        assert isinstance(output[0], Hypothesis)
        # check transcript
        assert output[0].text == 'stop'
        assert output[1].text == 'start'

        # check timestamp
        assert output[0].timestamp['segment'][0]['start'] == pytest.approx(0.4)
        assert output[0].timestamp['segment'][0]['end'] == pytest.approx(0.48)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_timestamps_with_transcribe_hybrid(self, audio_files, fast_conformer_hybrid_model):
        audio1, audio2 = audio_files

        output = fast_conformer_hybrid_model.transcribe([audio1, audio2], timestamps=True)

        # check len of output
        assert len(output) == 2

        # check hypothesis object
        assert isinstance(output[0], Hypothesis)
        # check transcript
        assert output[0].text == 'Stop?'
        assert output[1].text == 'Start.'

        # check timestamp
        assert output[0].timestamp['segment'][0]['start'] == pytest.approx(0.48)
        assert output[0].timestamp['segment'][0]['end'] == pytest.approx(0.72)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_timestamps_with_transcribe_hybrid_ctc_head(self, audio_files, fast_conformer_hybrid_model):
        audio1, audio2 = audio_files

        fast_conformer_hybrid_model.change_decoding_strategy(decoder_type='ctc')

        output = fast_conformer_hybrid_model.transcribe([audio1, audio2], timestamps=True)

        # check len of output
        assert len(output) == 2

        # check hypothesis object
        assert isinstance(output[0], Hypothesis)
        # check transcript
        assert output[0].text in ['Stop', 'Stop?']
        assert output[1].text in ['Start', 'Start.']

        # check timestamp
        assert output[0].timestamp['segment'][0]['start'] == pytest.approx(0.4)
        assert output[0].timestamp['segment'][0]['end'] == pytest.approx(0.72)

    @pytest.mark.with_downloads()
    @pytest.mark.unit
    def test_timestamps_with_transcribe_canary_flash(self, audio_files, canary_1b_flash):
        audio1, audio2 = audio_files

        output = canary_1b_flash.transcribe([audio1, audio2], timestamps=True)

        # check len of output
        assert len(output) == 2

        # check hypothesis object
        assert isinstance(output[0], Hypothesis)
        # check transcript
        assert output[0].text == 'Stop'
        assert output[1].text == 'start'

        # check timestamp
        assert output[0].timestamp['segment'][0]['start'] == pytest.approx(0.32)
        assert output[0].timestamp['segment'][0]['end'] == pytest.approx(0.72)

    @pytest.mark.unit
    def test_transcribe_return_nbest_hybrid_rnnt_ctc_prompt(self, audio_files, hybrid_rnnt_ctc_bpe_model_with_prompt):
        """Test n-best hypothesis return for hybrid RNNT-CTC BPE model with prompts."""
        hybrid_rnnt_ctc_bpe_model_with_prompt.eval()
        audio1, audio2 = audio_files

        orig_decoding_config = copy.deepcopy(hybrid_rnnt_ctc_bpe_model_with_prompt.cfg.decoding)

        decoding_config = copy.deepcopy(hybrid_rnnt_ctc_bpe_model_with_prompt.cfg.decoding)
        with open_dict(decoding_config):
            decoding_config["strategy"] = "beam"
            decoding_config["beam"]["beam_size"] = 4
            decoding_config["beam"]["return_best_hypothesis"] = False

        hybrid_rnnt_ctc_bpe_model_with_prompt.change_decoding_strategy(decoding_config)

        # Transcribe audio with prompt parameters
        hypotheses = hybrid_rnnt_ctc_bpe_model_with_prompt.transcribe(
            [audio1, audio2], batch_size=2, return_hypotheses=True, target_lang="en-US"
        )

        # Check results
        assert len(hypotheses) == 2
        assert isinstance(hypotheses[0], list)  # n-best list
        assert len(hypotheses[0]) > 0  # at least one hypothesis

        # Restore original decoding config
        hybrid_rnnt_ctc_bpe_model_with_prompt.change_decoding_strategy(orig_decoding_config)

    @pytest.mark.unit
    def test_timestamps_with_transcribe_hybrid_prompt(self, audio_files, hybrid_rnnt_ctc_bpe_model_with_prompt):
        audio1, audio2 = audio_files

        output = hybrid_rnnt_ctc_bpe_model_with_prompt.transcribe([audio1, audio2], timestamps=True)

        # check len of output
        assert len(output) == 2

        # check hypothesis object
        assert isinstance(output[0], Hypothesis)
        # check transcript
        assert output[0].text == 'Stop'
        assert output[1].text == 'Start'

        # check timestamp
        assert output[0].timestamp['segment'][0]['start'] == pytest.approx(0.16)
        assert output[0].timestamp['segment'][0]['end'] == pytest.approx(0.56)

    @pytest.mark.unit
    def test_transcribe_returns_xattn(self, audio_files, canary_1b_v2):
        canary_1b_v2.eval()
        audio1, audio2 = audio_files

        orig_decoding_config = copy.deepcopy(canary_1b_v2.cfg.decoding)

        decoding_config = MultiTaskDecodingConfig()
        decoding_config.return_xattn_scores = True
        canary_1b_v2.change_decoding_strategy(decoding_config)

        config = MultiTaskTranscriptionConfig(
            batch_size=4,
            return_hypotheses=True,
            num_workers=0,
            verbose=False,
            prompt={'source_lang': 'en', 'target_lang': 'en'},
            enable_chunking=False,
        )

        output = canary_1b_v2.transcribe([audio1, audio2], override_config=config)
        assert output[0].xatt_scores is not None
        assert output[1].xatt_scores is not None

        # Reset the decoding strategy to original
        canary_1b_v2.change_decoding_strategy(orig_decoding_config)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
from omegaconf import DictConfig

from nemo.core.classes import NeuralModule
from nemo.core.neural_types import LengthsType, LogprobsType, NeuralType, PredictionsType


class ViterbiDecoderWithGraph(NeuralModule):
    """Viterbi Decoder with WFSA (Weighted Finite State Automaton) graphs.

    Note:
        Requires k2 v1.14 or later to be installed to use this module.

    Decoder can be set up via the config, and optionally be passed keyword arguments as follows.

    Examples:
        .. code-block:: yaml

            model:  # Model config
                ...
                graph_module_cfg:  # Config for graph modules, e.g. ViterbiDecoderWithGraph
                    split_batch_size: 0
                    backend_cfg:
                        topo_type: "default"       # other options: "compact", "shared_blank", "minimal"
                        topo_with_self_loops: true
                        token_lm: <token_lm_path>  # must be provided for criterion_type: "map"

    Args:
        num_classes: Number of target classes for the decoder network to predict.
            (Excluding the blank token).

        backend: Which backend to use for decoding. Currently only `k2` is supported.

        dec_type: Type of decoding graph to use. Choices: `topo` and `token_lm`, 
            with `topo` standing for the loss topology graph only 
            and `token_lm` for the topology composed with a token_lm graph.

        return_type: Type of output. Choices: `1best` and `lattice`.
            `1best` is represented as a list of 1D tensors.
            `lattice` can be of type corresponding to the backend (e.g. k2.Fsa).

        return_ilabels: For return_type=`1best`.
            Whether to return input labels of a lattice (otherwise output labels).

        output_aligned: For return_type=`1best`.
            Whether the tensors length will correspond to log_probs_length 
            and the labels will be aligned to the frames of emission 
            (otherwise there will be only the necessary labels).

        split_batch_size: Local batch size. Used for memory consumption reduction at the cost of speed performance.
            Effective if complies 0 < split_batch_size < batch_size.

        graph_module_cfg: Optional Dict of (str, value) pairs that are passed to the backend graph decoder.
    """

    @property
    def input_types(self):
        """Returns definitions of module input ports.
        """
        return {
            "log_probs": NeuralType(("B", "T", "D") if self._3d_input else ("B", "T", "T", "D"), LogprobsType()),
            "input_lengths": NeuralType(tuple("B"), LengthsType()),
        }

    @property
    def output_types(self):
        """Returns definitions of module output ports.
        """
        return {"predictions": NeuralType(("B", "T"), PredictionsType())}

    def __init__(
        self,
        num_classes,
        backend: str = "k2",
        dec_type: str = "topo",
        return_type: str = "1best",
        return_ilabels: bool = True,
        output_aligned: bool = True,
        split_batch_size: int = 0,
        graph_module_cfg: Optional[DictConfig] = None,
    ):
        self._blank = num_classes
        self.return_ilabels = return_ilabels
        self.output_aligned = output_aligned
        self.split_batch_size = split_batch_size
        self.dec_type = dec_type

        if return_type == "1best":
            self.return_lattices = False
        elif return_type == "lattice":
            self.return_lattices = True
        elif return_type == "nbest":
            raise NotImplementedError(f"return_type {return_type} is not supported at the moment")
        else:
            raise ValueError(f"Unsupported return_type: {return_type}")

        # we assume that self._blank + 1 == num_classes
        if backend == "k2":
            if self.dec_type == "topo":
                from nemo.collections.asr.parts.k2.graph_decoders import CtcDecoder as Decoder
            elif self.dec_type == "topo_rnnt_ali":
                from nemo.collections.asr.parts.k2.graph_decoders import RnntAligner as Decoder
            elif self.dec_type == "token_lm":
                from nemo.collections.asr.parts.k2.graph_decoders import TokenLMDecoder as Decoder
            elif self.dec_type == "loose_ali":
                raise NotImplementedError()
            elif self.dec_type == "tlg":
                raise NotImplementedError(f"dec_type {self.dec_type} is not supported at the moment")
            else:
                raise ValueError(f"Unsupported dec_type: {self.dec_type}")

            self._decoder = Decoder(num_classes=self._blank + 1, blank=self._blank, cfg=graph_module_cfg)
        elif backend == "gtn":
            raise NotImplementedError("gtn-backed decoding is not implemented")

        self._3d_input = self.dec_type != "topo_rnnt"
        super().__init__()

    def update_graph(self, graph):
        """Updates graph of the backend graph decoder.
        """
        self._decoder.update_graph(graph)

    def _forward_impl(self, log_probs, log_probs_length, targets=None, target_length=None):
        if targets is None and target_length is not None or targets is not None and target_length is None:
            raise RuntimeError(
                f"Both targets and target_length have to be None or not None: {targets}, {target_length}"
            )
        # do not use self.return_lattices for now
        if targets is None:
            align = False
            decode_func = lambda a, b: self._decoder.decode(
                a, b, return_lattices=False, return_ilabels=self.return_ilabels, output_aligned=self.output_aligned
            )
        else:
            align = True
            decode_func = lambda a, b, c, d: self._decoder.align(
                a, b, c, d, return_lattices=False, return_ilabels=False, output_aligned=True
            )
        batch_size = log_probs.shape[0]
        if self.split_batch_size > 0 and self.split_batch_size <= batch_size:
            predictions = []
            probs = []
            for batch_idx in range(0, batch_size, self.split_batch_size):
                begin = batch_idx
                end = min(begin + self.split_batch_size, batch_size)
                log_probs_length_part = log_probs_length[begin:end]
                log_probs_part = log_probs[begin:end, : log_probs_length_part.max()]
                if align:
                    target_length_part = target_length[begin:end]
                    targets_part = targets[begin:end, : target_length_part.max()]
                    predictions_part, probs_part = decode_func(
                        log_probs_part, log_probs_length_part, targets_part, target_length_part
                    )
                    del targets_part, target_length_part
                else:
                    predictions_part, probs_part = decode_func(log_probs_part, log_probs_length_part)
                del log_probs_part, log_probs_length_part
                predictions += predictions_part
                probs += probs_part
        else:
            predictions, probs = (
                decode_func(log_probs, log_probs_length, targets, target_length)
                if align
                else decode_func(log_probs, log_probs_length)
            )
        assert len(predictions) == len(probs)
        return predictions, probs

    @torch.no_grad()
    def forward(self, log_probs, log_probs_length):
        if self.dec_type == "looseali":
            raise RuntimeError(f"Decoder with dec_type=`{self.dec_type}` is not intended for regular decoding.")
        predictions, probs = self._forward_impl(log_probs, log_probs_length)
        lengths = torch.tensor([len(pred) for pred in predictions], device=predictions[0].device)
        predictions_tensor = torch.full((len(predictions), lengths.max()), self._blank).to(
            device=predictions[0].device
        )
        probs_tensor = torch.full((len(probs), lengths.max()), 1.0).to(device=predictions[0].device)
        for i, (pred, prob) in enumerate(zip(predictions, probs)):
            predictions_tensor[i, : lengths[i]] = pred
            probs_tensor[i, : lengths[i]] = prob
        return predictions_tensor, lengths, probs_tensor

    @torch.no_grad()
    def align(self, log_probs, log_probs_length, targets, target_length):
        len_enough = (log_probs_length >= target_length) & (target_length > 0)
        if torch.all(len_enough) or self.dec_type == "looseali":
            results = self._forward_impl(log_probs, log_probs_length, targets, target_length)
        else:
            results = self._forward_impl(
                log_probs[len_enough], log_probs_length[len_enough], targets[len_enough], target_length[len_enough]
            )
            for i, computed in enumerate(len_enough):
                if not computed:
                    results[0].insert(i, torch.empty(0, dtype=torch.int32))
                    results[1].insert(i, torch.empty(0, dtype=torch.float))
        return results



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo.collections.common.losses.aggregator import AggregatorLoss
from nemo.collections.common.losses.bce_logits_loss import BCEWithLogitsLoss
from nemo.collections.common.losses.cross_entropy import CrossEntropyLoss, NLLLoss
from nemo.collections.common.losses.mse_loss import MSELoss
from nemo.collections.common.losses.multi_similarity_loss import MultiSimilarityLoss
from nemo.collections.common.losses.smoothed_cross_entropy import SmoothedCrossEntropyLoss, SmoothedNLLLoss
from nemo.collections.common.losses.spanning_loss import SpanningLoss



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test NeMo 2.0 local checkpointing functionality with NVRx resiliency extensions.

This test verifies the ability to save local checkpoints during training, simulate
a crash, and then successfully resume training from the latest available local
checkpoint, ensuring the trainer state (specifically global_step) is correctly restored.

The test proceeds in two main runs:

Run 1:
  - Initializes a LlamaModel and trains it using a MockDataModule.
  - Uses `LocalCheckpointCallback` from NVRx to save checkpoints locally at a
    specified interval (`--local-checkpoint-interval`).
  - Uses `CrashCallback` to simulate a crash by raising an exception at a
    predefined step (`--crash-step`).
  - After the simulated crash, the test verifies that the latest local checkpoint
    found on the node corresponds to the expected step based on the crash step
    and the local checkpoint interval.

Run 2:
  - Re-initializes the model, optimizer, and trainer.
  - Uses `nl.AutoResume` to automatically detect and load the latest checkpoint
    (which should be the local checkpoint saved just before the crash in Run 1).
  - Includes `CheckResumeStepCallback` which asserts, at the `on_train_start`
    hook, that `trainer.global_step` matches the step number of the local
    checkpoint from which it resumed. This confirms state restoration.
  - Continues training until `--max-steps`.
  - Finally, verifies that the latest local checkpoint after Run 2 completes
    corresponds to the expected final step based on `max_steps` and the local
    checkpoint interval.

Requirements:
  - nvidia_resiliency_ext (`res_module`) must be installed.
  - A shared filesystem for `--log-dir` if running on multiple nodes.
"""

import argparse
import logging
import os
import re
import shutil
import socket
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import nemo_run as run
import torch
import torch.distributed as dist
from lightning.pytorch.callbacks import Callback
from megatron.core.distributed import DistributedDataParallelConfig
from megatron.core.optimizer import OptimizerConfig

from nemo import lightning as nl
from nemo.collections import llm
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.gpt.model.llama import Llama3Config, LlamaModel
from nemo.collections.llm.recipes.log.default import get_global_step_from_global_checkpoint_path
from nemo.lightning.pytorch.callbacks import ModelCheckpoint
from nemo.lightning.pytorch.local_ckpt import update_trainer_local_checkpoint_io
from nemo.lightning.pytorch.optim.megatron import MegatronOptimizerModule
from nemo.utils.import_utils import safe_import

logger = logging.getLogger(__name__)

res_module, HAVE_RES = safe_import("nvidia_resiliency_ext.ptl_resiliency")


@dataclass
class Llama3Config145M(Llama3Config):
    rotary_base: int = 500_000
    seq_length: int = 8192  # Reduced for faster testing if needed
    num_layers: int = 2
    hidden_size: int = 768
    ffn_hidden_size: int = 2688
    num_attention_heads: int = 16


class CrashException(Exception):
    """Custom exception for triggering simulated crashes in tests."""

    pass


class CrashCallback(Callback):
    """
    Callback to simulate a crash at a specific step based on trainer.global_step.
    """

    def __init__(self, crash_step: int):
        if crash_step <= 0:
            raise ValueError("crash_step must be a positive integer or None")
        self.crash_step = crash_step
        if self.crash_step:
            logger.debug(f"CrashCallback initialized. Will simulate crash if global_step reaches {self.crash_step}")
        else:
            logger.debug("CrashCallback initialized. Crash simulation is disabled (crash_step is None).")

    def on_train_batch_end(self, trainer, *args) -> None:
        current_global_step = trainer.global_step
        if current_global_step == self.crash_step:
            msg = f"Simulating crash via CrashCallback at global_step {current_global_step}!"
            logger.error(msg)
            raise CrashException(msg)


class CheckResumeStepCallback(Callback):
    """
    Callback to verify the trainer's global_step at the start of training.
    Used to ensure resumption from a checkpoint happened correctly.
    """

    def __init__(self, expected_resume_step: int):
        self.expected_resume_step = expected_resume_step

    def on_train_start(self, trainer, *args) -> None:
        # We expect global_step to be the same as the step of the checkpoint
        # because training resumes from the saved step state.
        expected_step_at_start = self.expected_resume_step
        current_global_step = trainer.global_step

        assert (
            current_global_step == expected_step_at_start
        ), f"""Resumption check failed!
            Expected trainer.global_step to be {expected_step_at_start} after resuming, 
            but found {current_global_step}.
            """


def get_trainer(args, callbacks, plugins, strategy) -> nl.Trainer:
    """Creates a PyTorch Lightning Trainer instance."""
    return nl.Trainer(
        accelerator="gpu",
        devices=args.devices,
        num_nodes=args.num_nodes,
        max_steps=args.max_steps,
        callbacks=callbacks,
        log_every_n_steps=10,
        val_check_interval=args.val_check_interval,
        limit_val_batches=args.limit_val_batches,
        plugins=plugins,
        strategy=strategy,
        enable_progress_bar=False,
    )


def get_megatron_strategy(args, async_save: bool = True) -> nl.MegatronStrategy:
    """Creates a NeMo MegatronStrategy instance."""
    return nl.MegatronStrategy(
        tensor_model_parallel_size=1,
        pipeline_model_parallel_size=1,
        pipeline_dtype=None,
        virtual_pipeline_model_parallel_size=None,
        context_parallel_size=1,
        sequence_parallel=False,
        ckpt_async_save=async_save,
        ckpt_parallel_load=False,
        ddp=DistributedDataParallelConfig(),
    )


def get_optimizer(bf16_enabled: bool = True) -> MegatronOptimizerModule:
    """Creates a MegatronOptimizerModule instance."""
    opt_config = OptimizerConfig(
        optimizer='adam',
        lr=1e-4,
        weight_decay=0.1,
        adam_beta1=0.9,
        adam_beta2=0.95,
        adam_eps=1e-8,
        clip_grad=1.0,
        log_num_zeros_in_grad=False,
        timers=None,
        bf16=bf16_enabled,
        use_distributed_optimizer=False,
    )
    return MegatronOptimizerModule(config=opt_config)


def get_my_local_ckpt_node_dir(log_dir: str, rank: int) -> Path:
    """Gets the expected local checkpoint directory path."""
    return Path(log_dir) / "local_ckpt" / socket.gethostname() / str(rank)


def find_latest_local_ckpt_step(local_ckpt_node_dir: Path, global_rank: int) -> Optional[int]:
    """Finds the step number of the latest local checkpoint based on the input path."""
    latest_step = -1

    if not local_ckpt_node_dir.is_dir():
        logger.warning(f"Local checkpoint node directory not found: {local_ckpt_node_dir}")
        return None

    # Pattern to match the local checkpoint file name
    pattern = re.compile(fr"iter_(\d+)_{global_rank}_local")

    for item in local_ckpt_node_dir.iterdir():
        match = pattern.match(item.name)
        if match:
            step = int(match.group(1))
            if step > latest_step:
                latest_step = step

    result = latest_step if latest_step != -1 else None
    return result


def get_parser() -> argparse.ArgumentParser:
    """Creates the argument parser."""
    parser = argparse.ArgumentParser(description="Llama3 Local Ckpt Crash Test")
    parser.add_argument("--log-dir", type=str, required=True, help="Filesystem output dir.")
    parser.add_argument("--num-nodes", type=int, default=1, help="Total nodes in the job (usually from SLURM)")
    parser.add_argument("--devices", type=int, default=2, help="GPUs per node (usually from SLURM)")
    parser.add_argument('--max-steps', type=int, default=200, help="Total steps for training")
    parser.add_argument('--checkpoint-interval', type=int, default=80, help="Global checkpoint interval")
    parser.add_argument('--local-checkpoint-interval', type=int, default=45, help="Local checkpoint interval")
    parser.add_argument('--val-check-interval', type=int, default=80, help="Validation interval")
    parser.add_argument('--limit_val_batches', type=int, default=10, help="Validation batches limit")
    parser.add_argument("--async-save", action="store_true", help="Use async global ckpt save")
    parser.add_argument("--crash-step", type=int, default=100, help="Global step for Rank 0 to simulate crash")
    parser.add_argument("--cleanup-log-dir", action="store_true", help="Rank 0 cleans up log dir before starting")
    parser.add_argument(
        "--log-level",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level",
    )
    return parser


def main() -> None:
    args = get_parser().parse_args()

    log_level_numeric = getattr(logging, args.log_level.upper(), logging.INFO)
    logging.basicConfig(
        level=log_level_numeric,
        format='%(asctime)s - %(levelname)s - %(message)s',
        stream=sys.stdout,  # Log to standard output
    )

    assert HAVE_RES, "nvidia_resiliency_ext is required for this test."

    assert (
        args.crash_step > args.local_checkpoint_interval
    ), "Crash step must be after the first local checkpoint interval"
    assert args.crash_step < args.max_steps, "Crash step must be before max_steps"
    assert args.local_checkpoint_interval > 0, "Local checkpoint interval must be positive"

    executor = run.LocalExecutor(ntasks_per_node=args.devices, launcher="torchrun")
    # Convert Namespace to dict for serialization
    args_dict = vars(args)
    script = run.Partial(run_test, args_dict)
    run.run(script, executor)


def run_test(args_dict: dict) -> None:
    # Convert dict back to Namespace for compatibility with existing code
    args = argparse.Namespace(**args_dict)

    global_rank = int(os.environ.get("RANK", 0))
    world_size = int(os.environ.get("WORLD_SIZE", 1))

    if global_rank == 0:
        logger.debug(f"Test started. World size: {world_size}")
        logger.debug(f"Using shared log directory: {args.log_dir}")
        logger.debug(f"Logging level set to: {args.log_level.upper()}")
        if args.crash_step > 0:
            logger.debug(f"Crash simulation enabled at step {args.crash_step}.")
        assert (
            Path(args.log_dir).exists() or args.cleanup_log_dir
        ), f"Log directory {args.log_dir} does not exist and cleanup was not requested."

    # Log Directory Handling (Rank 0 Only)
    log_dir_path = Path(args.log_dir)
    if args.cleanup_log_dir and global_rank == 0:
        if log_dir_path.exists():
            shutil.rmtree(log_dir_path)
        logger.debug(f"Creating log directory: {log_dir_path}")
        log_dir_path.mkdir(parents=True, exist_ok=True)

    if torch.distributed.is_available() and torch.distributed.is_initialized():
        dist.barrier()

    my_local_ckpt_node_dir = get_my_local_ckpt_node_dir(args.log_dir, global_rank)
    logger.debug(f"Expecting local checkpoints for this node in: {my_local_ckpt_node_dir}")

    mbs = 1
    gbs = mbs * world_size
    model_config = Llama3Config145M()
    data = MockDataModule(seq_length=model_config.seq_length, global_batch_size=gbs, micro_batch_size=mbs)
    precision_plugin = nl.MegatronMixedPrecision(precision="bf16-mixed")
    strategy = get_megatron_strategy(args, async_save=args.async_save)

    model_run1 = LlamaModel(config=model_config)
    optim_run1 = get_optimizer(bf16_enabled=True)

    checkpoint_callback = ModelCheckpoint(
        save_last=True,
        monitor="val_loss",
        save_top_k=1,
        every_n_train_steps=args.checkpoint_interval,
        save_on_train_epoch_end=True,
        save_optim_on_train_end=True,
        filename='{model_name}--{val_loss:.2f}-{step}-{consumed_samples}',
    )
    local_checkpoint_callback = res_module.local_checkpoint_callback.LocalCheckpointCallback(
        every_n_train_steps=args.local_checkpoint_interval
    )

    # Instantiate the CrashCallback for Run 1 if crash_step is positive
    crash_callback = CrashCallback(crash_step=args.crash_step if args.crash_step > 0 else None)

    # Define the callback list for Run 1
    # Place LocalCheckpointCallback before CrashCallback
    callbacks_run1 = [
        local_checkpoint_callback,
        crash_callback,
    ]

    nemo_logger_plugin = nl.NeMoLogger(
        log_dir=args.log_dir,
        wandb=None,
        ckpt=checkpoint_callback,
    )

    trainer_run1 = get_trainer(
        args,
        callbacks=callbacks_run1,
        plugins=[precision_plugin],
        strategy=strategy,
    )
    assert hasattr(trainer_run1, 'global_rank'), "Trainer needs global_rank populated."

    logger.debug(f"Initializing Run 1 Trainer complete. Patching for local I/O.")
    update_trainer_local_checkpoint_io(
        trainer_run1,
        args.log_dir,
        get_global_step_from_global_checkpoint_path,
    )

    resume_logic = nl.AutoResume(resume_if_exists=True, resume_ignore_no_checkpoint=True)
    crashed = False
    # Run 1: Train until Crash
    if global_rank == 0:
        # Use logger
        logger.info("\n" + "=" * 20 + " Starting Run 1: Train until crash " + "=" * 20)

    try:
        llm.train(
            model=model_run1,
            data=data,
            trainer=trainer_run1,
            log=nemo_logger_plugin,
            optim=optim_run1,
            resume=resume_logic,
            tokenizer="data",
        )
    except CrashException as e:
        logger.debug(f"\nSuccessfully caught expected crash: {e}")
        crashed = True

    if torch.distributed.is_initialized():
        dist.barrier()

    latest_local_step_run1 = -1

    logger.debug("\nVerifying checkpoints after Run 1...")
    assert crashed, "Training did not crash as expected!"
    logger.debug(f"Run 1 finished (crashed at step {args.crash_step}).")

    latest_local_step_run1 = find_latest_local_ckpt_step(my_local_ckpt_node_dir, global_rank)
    assert latest_local_step_run1 is not None, f"No local checkpoints found in {my_local_ckpt_node_dir} after crash!"

    expected_latest_step = (args.crash_step // args.local_checkpoint_interval) * args.local_checkpoint_interval
    logger.debug(f"Expected latest local checkpoint step: {expected_latest_step}")
    assert (
        latest_local_step_run1 == expected_latest_step
    ), f"Latest local step {latest_local_step_run1} is not equal to the expected {expected_latest_step}"

    logger.debug(f"Latest local checkpoint found at step: {latest_local_step_run1}")

    # --- Run 2: Resume Training ---
    if global_rank == 0:
        # Use logger
        logger.info(
            "\n" + "=" * 20 + f" Starting Run 2: All ranks resuming from step {latest_local_step_run1} " + "=" * 20
        )

    if torch.distributed.is_initialized():
        dist.barrier()

    data = MockDataModule(seq_length=model_config.seq_length, global_batch_size=gbs, micro_batch_size=mbs)
    precision_plugin_run2 = nl.MegatronMixedPrecision(precision="bf16-mixed")
    strategy_run2 = get_megatron_strategy(args, async_save=args.async_save)

    model_run2 = LlamaModel(config=model_config)
    optim_run2 = get_optimizer(bf16_enabled=True)
    data = MockDataModule(seq_length=model_config.seq_length, global_batch_size=gbs, micro_batch_size=mbs)

    checkpoint_callback_run2 = ModelCheckpoint(
        save_last=True,
        monitor="val_loss",
        save_top_k=1,
        every_n_train_steps=args.checkpoint_interval,
        save_on_train_epoch_end=True,
        save_optim_on_train_end=True,
        filename='{model_name}--{val_loss:.2f}-{step}-{consumed_samples}',
    )
    local_checkpoint_callback_run2 = res_module.local_checkpoint_callback.LocalCheckpointCallback(
        every_n_train_steps=args.local_checkpoint_interval
    )
    # Instantiate the check callback with the expected resume step
    check_resume_step_callback = CheckResumeStepCallback(expected_resume_step=latest_local_step_run1)
    callbacks_run2 = [local_checkpoint_callback_run2, check_resume_step_callback]

    nemo_logger_plugin_run2 = nl.NeMoLogger(
        log_dir=args.log_dir,
        wandb=None,
        ckpt=checkpoint_callback_run2,
    )

    trainer_run2 = get_trainer(
        args,
        callbacks=callbacks_run2,
        plugins=[precision_plugin_run2],
        strategy=strategy_run2,
    )

    assert hasattr(trainer_run2, 'global_rank'), "Trainer needs global_rank populated."
    assert trainer_run2.global_rank == global_rank, "Trainer rank differs from env var rank!"
    assert trainer_run2.world_size == world_size, "Trainer world size differs from env var world size!"

    update_trainer_local_checkpoint_io(
        trainer_run2,
        args.log_dir,
        get_global_step_from_global_checkpoint_path,
    )

    resume_logic_run2 = nl.AutoResume(
        resume_if_exists=True,
        resume_ignore_no_checkpoint=False,
    )
    logger.debug(f"[Rank {global_rank}] Starting Run 2 training loop (resuming).")
    llm.train(
        model=model_run2,
        data=data,
        trainer=trainer_run2,
        log=nemo_logger_plugin_run2,
        optim=optim_run2,
        resume=resume_logic_run2,
        tokenizer="data",
    )
    logger.debug(f"Run 2 resumed from step {latest_local_step_run1}.")
    latest_local_step_run2 = find_latest_local_ckpt_step(my_local_ckpt_node_dir, global_rank)

    assert (
        latest_local_step_run2 is not None
    ), f"No local checkpoints found in {my_local_ckpt_node_dir} after resuming!"

    expected_latest_step = (args.max_steps // args.local_checkpoint_interval) * args.local_checkpoint_interval
    logger.debug(f"Expected latest local checkpoint step: {expected_latest_step}")
    assert (
        latest_local_step_run2 == expected_latest_step
    ), f"Latest local step {latest_local_step_run2} is not equal to the expected {expected_latest_step}"
    logger.debug(f"Latest local checkpoint found at step: {latest_local_step_run2}")

    if torch.distributed.is_initialized():
        dist.barrier()


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest

from nemo.collections.llm.api import finetune
from nemo.collections.vlm import Llava15Config13B, LlavaModel, LoRA
from nemo.collections.vlm.recipes import llava15_13b
from nemo.lightning import Trainer
from nemo.lightning.pytorch.callbacks.megatron_comm_overlap import MegatronCommOverlapCallback
from nemo.utils.exp_manager import TimingCallback


class TestLlava15_13B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        # The new recipe module is assumed to be imported as llava15_13b
        return llava15_13b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        # Verify that model() returns a run.Config for the LlavaModel
        assert isinstance(model_config, run.Config)
        assert model_config.__fn_or_cls__ == LlavaModel
        # Check that the inner config is set to Llava15Config13B
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == Llava15Config13B

    def test_finetune_recipe_default(self, recipe_module):
        # Default peft_scheme is 'lora'
        recipe = recipe_module.finetune_recipe()
        # Check the recipe is a run.Partial wrapping the finetune function
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune

        # Check that the model configuration is correct
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == LlavaModel

        # Check trainer configuration
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert recipe.trainer.accelerator == "gpu"
        assert recipe.trainer.devices == 8
        assert recipe.trainer.num_nodes == 1
        # Check some trainer parameters
        assert recipe.trainer.max_steps == 5190
        # Validate strategy configuration
        strat = recipe.trainer.strategy
        assert isinstance(strat, run.Config)
        # By default for 'lora', the recipe should not change tensor parallelism in the "none" branch.
        # So we expect the default strategy to remain (tensor_model_parallel_size initially 1)
        # but note that when using LoRA, the branch does not update tensor parallelism.
        # Instead, we validate that the PEFT configuration is applied.
        assert hasattr(recipe, "peft")
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == LoRA
        # For LoRA, learning rate is updated
        assert recipe.optim.config.lr == 1e-4

        # Verify data configuration
        assert isinstance(recipe.data, run.Config)
        # Check that we are using the MockDataModule with expected parameters
        assert recipe.data.__fn_or_cls__.__name__ == "MockDataModule"
        assert recipe.data.seq_length == 4096
        assert recipe.data.global_batch_size == 128
        # For this recipe, micro_batch_size should be 1
        assert recipe.data.micro_batch_size == 1
        assert recipe.data.num_workers == 4

        # Validate logging and resume configuration are set
        assert recipe.log is not None
        assert recipe.resume is not None

        # Validate trainer callbacks contain TimingCallback and MegatronCommOverlapCallback
        callback_classes = {cb.__fn_or_cls__ for cb in recipe.trainer.callbacks if isinstance(cb, run.Config)}
        assert TimingCallback in callback_classes
        assert MegatronCommOverlapCallback in callback_classes

    def test_finetune_recipe_peft_none(self, recipe_module):
        # Test the case where peft_scheme is set to 'none'
        recipe = recipe_module.finetune_recipe(peft_scheme="none")
        # In this branch, the recipe should not add a PEFT configuration
        assert hasattr(recipe, "peft")
        # And the strategy should be updated to use tensor_model_parallel_size = 2
        assert recipe.trainer.strategy.tensor_model_parallel_size == 2
        # And learning rate should be set to a lower value
        assert recipe.optim.config.lr == 2e-05

    @pytest.mark.parametrize("num_nodes,num_gpus", [(1, 8), (2, 4), (4, 2)])
    def test_parameterized_configurations(self, recipe_module, num_nodes, num_gpus):
        recipe = recipe_module.finetune_recipe(num_nodes=num_nodes, num_gpus_per_node=num_gpus)
        assert recipe.trainer.num_nodes == num_nodes
        assert recipe.trainer.devices == num_gpus



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any

import torch
from lhotse import fastcopy
from lhotse.serialization import SequentialJsonlWriter
from lightning import LightningModule
from omegaconf import DictConfig, open_dict
from peft import PeftModel
from torch import Tensor
from torch.distributed.fsdp import fully_shard, register_fsdp_forward_method
from torch.distributed.tensor import Replicate, Shard
from torch.distributed.tensor.parallel import (
    ColwiseParallel,
    PrepareModuleInput,
    RowwiseParallel,
    SequenceParallel,
    loss_parallel,
    parallelize_module,
)
from transformers import GenerationConfig

from nemo.collections.common.data.lhotse import NeMoMultimodalConversation
from nemo.collections.common.data.lhotse.dataloader import tokenize_with_prompt
from nemo.collections.common.data.lhotse.text_adapters import TextTurn
from nemo.collections.common.prompts import PromptFormatter
from nemo.collections.common.tokenizers import AutoTokenizer
from nemo.collections.speechlm2.data.salm_dataset import left_collate_vectors
from nemo.collections.speechlm2.models.salm import _resolve_audios_in_prompt, replace_placeholders_and_build_targets
from nemo.collections.speechlm2.modules.perception import AudioTranscriptionPerceptionModule
from nemo.collections.speechlm2.parts.hf_hub import HFHubMixin
from nemo.collections.speechlm2.parts.lora import maybe_install_lora
from nemo.collections.speechlm2.parts.optim_setup import configure_optimizers, is_frozen
from nemo.collections.speechlm2.parts.pretrained import load_pretrained_hf, move_embedding
from nemo.core.neural_types import AudioSignal, LabelsType, LengthsType, MaskType, NeuralType
from nemo.utils import logging


class SALMWithAsrDecoder(LightningModule, HFHubMixin):
    def __init__(self, cfg) -> None:
        assert isinstance(cfg, dict), (
            "You must pass the config to SALM as a Python dict to support hyperparameter serialization "
            f"in PTL checkpoints (we got: '{type(cfg)=}')."
        )
        super().__init__()
        self.save_hyperparameters()
        self.cfg = DictConfig(cfg)
        self.audio_locator_tag = self.cfg.audio_locator_tag

        self.tokenizer = AutoTokenizer(self.cfg.pretrained_llm, use_fast=True)
        self.tokenizer.add_special_tokens({"additional_special_tokens": [self.audio_locator_tag]})
        self.llm = load_pretrained_hf(self.cfg.pretrained_llm, pretrained_weights=self.cfg.pretrained_weights)
        if not hasattr(self.llm, "model") and hasattr(self.llm, "backbone"):
            type(self.llm).model = property(lambda self: self.backbone)
        if not hasattr(self.llm.model, "embed_tokens") and hasattr(self.llm.model, "embeddings"):
            self.llm.model.embed_tokens = self.llm.model.embeddings
        # Note: we have to "move out" the token embedding outside of LLM to avoid
        #       messing up FSDP/TP hooks.
        self.embed_tokens = self.llm.model.embed_tokens
        del self.llm.model.embed_tokens

        # Load the pretrained streaming ASR model and copy its parameters into the audio perception module.
        setup_speech_encoder_with_asr(self, pretrained_weights=self.cfg.pretrained_weights)
        assert isinstance(self.perception, AudioTranscriptionPerceptionModule)

        # Load pretrained weights if provided
        if (init_from_path := self.cfg.get("init_from_path", None)) is not None:
            init_from_path = Path(init_from_path)
            assert init_from_path.is_dir(), "init_from_path must be a directory containing HF checkpoint"
            logging.warning(f"Loading pretrained weights from {str(init_from_path)}")
            from safetensors import safe_open

            tensors = {}
            with safe_open(init_from_path / "model.safetensors", framework="pt") as f:
                for k in f.keys():
                    tensors[k] = f.get_tensor(k)
            missing_keys, unexpected_keys = self.load_state_dict(tensors, strict=False)
            logging.warning(f"Missing keys: {missing_keys}")
            logging.warning(f"Unexpected keys: {unexpected_keys}")

        maybe_install_lora(self)

        self._use_fsdp = False
        self._use_tp = False

    @property
    def text_vocab_size(self):
        """Return the size of the text tokenizer."""
        return self.embed_tokens.num_embeddings

    @property
    def text_bos_id(self) -> int:
        return self.tokenizer.bos_id

    @property
    def text_eos_id(self) -> int:
        return self.tokenizer.eos_id

    @property
    def text_pad_id(self) -> int:
        pad_id = self.tokenizer.pad
        if pad_id is None:
            pad_id = self.tokenizer.unk_id
        if pad_id is None:
            warnings.warn(
                "the text tokenizer has no <pad> or <unk> tokens available, using id 0 for padding (this may lead to silent bugs)."
            )
            pad_id = 0
        return pad_id

    @property
    def audio_locator_tag_id(self) -> int:
        return self.tokenizer.token_to_id(self.audio_locator_tag)

    @property
    def token_equivalent_duration(self) -> float:
        """
        Returns the audio duration corresponding to a single frame/token at the output of ``self.perception``.
        """
        return self.perception.token_equivalent_duration

    @property
    def sampling_rate(self) -> int:
        return self.perception.preprocessor.featurizer.sample_rate

    def forward(
        self,
        input_embeds: Tensor,
        attention_mask: Tensor = None,
        cache=None,
    ) -> dict[str, Tensor]:
        """
        Implements a fully offline forward pass through the entire model.
        The flow is the following:

        |speech and text embeddings| -> |llm| -> |lm_head| -> |token ids|

        """
        # input_embeds and out: (B, T, H)
        out = self.llm(
            inputs_embeds=input_embeds,
            attention_mask=attention_mask,
            past_key_values=cache,
            use_cache=cache is not None,
            return_dict=True,
        )
        ans = {"logits": out['logits']}  # (B, T, text_vocab_size)
        if cache is not None:
            ans["cache"] = out["past_key_values"]
        return ans

    def prepare_inputs(self, batch: dict):
        """
        Performs additional processing on the mini-batch collected from dataloader.
        Notably:
        * Convert source audio to speech representations.
        * Convert target audio to target audio tokens.
        * Convert target text to embeddings.
        * Combine the input audio and target text embeddings.
        * Take care of any necessary slicing to align the shapes of source audio,
            target audio, and target token ids.
        """
        # Source audio encoding.
        # Input audio: (B, T_samples)
        # Audio embeddings: (B, T, H)
        encoded, encoded_len = self.perception.forward_encoder(
            input_signal=batch["audios"], input_signal_length=batch["audio_lens"]
        )
        asr_hyps = self.perception.transcribe_encoded(encoded=encoded, encoded_len=encoded_len)
        # During training, we randomly drop the transcript
        for hyp in asr_hyps:
            if self.training and random.random() < self.cfg.get("asr_transcript_drop_prob", 0.0):
                hyp.text = ""
        asr_tokens = [
            torch.as_tensor(self.tokenizer.text_to_ids(f">> {hyp.text} <<" if hyp.text else ">> <<"))
            for hyp in asr_hyps
        ]
        asr_tokens_len = [at.shape[0] for at in asr_tokens]
        asr_tokens = torch.cat(asr_tokens, dim=0).unsqueeze(0).to(self.device)
        transcript_embs = torch.split(self.embed_tokens(asr_tokens).squeeze(0), asr_tokens_len, dim=0)
        audio_embs, audio_emb_lens = self.perception(encoded=encoded, encoded_len=encoded_len)
        audio_embs = [
            torch.cat([aemb[:aemblen], temb], dim=0)
            for aemb, aemblen, temb in zip(audio_embs, audio_emb_lens, transcript_embs)
        ]
        input_ids_to_embed = torch.where(batch["input_ids"] == self.audio_locator_tag_id, 0, batch["input_ids"])
        text_embs = self.embed_tokens(input_ids_to_embed)
        input_embs, target_ids, attention_mask = replace_placeholders_and_build_targets(
            input_ids=batch["input_ids"],
            embeds=text_embs,
            padding_id=self.text_pad_id,
            placeholder_id=self.audio_locator_tag_id,
            replacements=audio_embs,
            target_ids=batch["input_ids"].where(batch["loss_mask"], -100),  # CrossEntropyLoss().ignore_index
        )
        input_embs = input_embs[:, :-1]
        attention_mask = attention_mask[:, :-1]
        target_ids = target_ids[:, 1:]

        # Combine target audio and text into a single tensor to slice them together.
        # It will also help us truncate the sequence lengths to be divisible by TP world size,
        # when TP is enabled.
        # Input ids: (B, T, K+1)
        if self._use_tp:
            tp_world_size = self.device_mesh["tensor_parallel"].size()
            if (remainder := (input_embs.shape[1] - 1) % tp_world_size) != 0:
                # Truncate some tokens from the end to make the sequence lenght shape divisible by tensor parallelism
                # world size. Otherwise, sequence parallelism will change the input shape making leading to mismatches.
                input_embs = input_embs[:, :-remainder]
                attention_mask = attention_mask[:, :-remainder]
                target_ids = target_ids[:, :-remainder]

        return {
            "input_embeds": input_embs,
            "attention_mask": attention_mask,
            "target_ids": target_ids,
        }

    def training_step(self, batch: dict, batch_idx: int):
        for m in (self.perception.preprocessor, self.perception.encoder, self.llm):
            if is_frozen(m):
                m.eval()

        inputs = self.prepare_inputs(batch)
        forward_outputs = self(inputs["input_embeds"], attention_mask=inputs["attention_mask"])
        num_frames = (inputs["target_ids"] != -100).long().sum()
        with loss_parallel():
            loss = (
                torch.nn.functional.cross_entropy(
                    forward_outputs["logits"].flatten(0, 1),  # (B, T, Vt) -> (*, Vt)
                    inputs["target_ids"].flatten(0, 1),
                    reduction="sum",
                    ignore_index=-100,
                )
                / num_frames
            )

        B, T = inputs["input_embeds"].shape[:2]
        ans = {
            "loss": loss,
            "learning_rate": (
                torch.as_tensor(self.trainer.optimizers[0].param_groups[0]['lr'] if self._trainer is not None else 0)
            ),
            "batch_size": B,
            "sequence_length": T,
            "num_frames": num_frames.to(torch.float32),  # avoid warning
            "target_to_input_ratio": num_frames / (B * T),
            "padding_ratio": (batch["input_ids"] != self.text_pad_id).long().sum() / batch["input_ids"].numel(),
        }
        self.log_dict(ans, on_step=True)
        return ans

    def on_validation_epoch_start(self) -> None:
        self._partial_val_losses = defaultdict(list)
        self._partial_accuracies = defaultdict(list)

        # collect generations per validation set (per-rank)
        self._val_generations = defaultdict(list)

    def on_validation_epoch_end(self) -> None:
        val_losses = []
        for name, vals in self._partial_val_losses.items():
            val_loss = torch.stack(vals).mean()
            self.log(f"val_loss_{name}", val_loss, on_epoch=True, sync_dist=True)
            val_losses.append(val_loss)
        self.log("val_loss", torch.stack(val_losses).mean(), on_epoch=True, sync_dist=True)

        accuracies = []
        for name, accs in self._partial_accuracies.items():
            val_acc = torch.stack(accs).mean()
            self.log(f"val_acc_{name}", val_acc, on_epoch=True, sync_dist=True)
            accuracies.append(val_acc)
        self.log("val_acc", torch.stack(accuracies).mean(), on_epoch=True, sync_dist=True)

        self._partial_val_losses.clear()
        self._partial_accuracies.clear()

        # Gather and write generations to a single file per dataset (rank 0 only)
        if self.cfg.get("val_save_path", None) is not None:
            dist = torch.distributed
            if dist.is_available() and dist.is_initialized():
                world_size = dist.get_world_size()
                gathered = [None for _ in range(world_size)]
                dist.all_gather_object(gathered, dict(self._val_generations))
                is_global_zero = dist.get_rank() == 0
            else:
                gathered = [dict(self._val_generations)]
                is_global_zero = True

            if is_global_zero:
                merged = defaultdict(list)
                for per_rank_dict in gathered:
                    for name, items in per_rank_dict.items():
                        merged[name].extend(items)

                val_save_path = Path(self.cfg.val_save_path) / f"{self.global_step:06d}"
                val_save_path.mkdir(parents=True, exist_ok=True)
                for name, items in merged.items():
                    out_path = val_save_path / f"{name}.jsonl"
                    with SequentialJsonlWriter(out_path) as writer:
                        for obj in items:
                            writer.write(obj)

        self._val_generations.clear()

    def validation_step(self, batch: dict, batch_idx: int):
        for name, dataset_batch in batch.items():
            if dataset_batch is None:
                continue  # some dataset is exhausted

            try:
                inputs = self.prepare_inputs(dataset_batch)
                forward_outputs = self(inputs["input_embeds"], attention_mask=inputs["attention_mask"])
                num_frames = (inputs["target_ids"] != -100).long().sum()
                with loss_parallel():
                    loss = (
                        torch.nn.functional.cross_entropy(
                            forward_outputs["logits"].flatten(0, 1),
                            inputs["target_ids"].flatten(0, 1),
                            reduction="sum",
                            ignore_index=-100,
                        )
                        / num_frames
                    )

                preds = forward_outputs["logits"].argmax(dim=-1).view(-1)
                refs = inputs["target_ids"].reshape(-1)
                preds = preds[refs != -100]
                refs = refs[refs != -100]
                accuracy = preds.eq(refs).float().mean()

                self._partial_accuracies[name].append(accuracy)
                self._partial_val_losses[name].append(loss)

            except Exception as e:
                # Skip the dataset if there is an error, e.g., the dataset does not have answers
                logging.warning_once(f"Error in validation step for dataset {name}: {e}")

            # Run autoregressive generation and collect results (writing happens at epoch end)
            if self.cfg.get("val_save_path", None) is not None:
                convs_no_answer = [strip_response_if_any(conv) for conv in dataset_batch["conversations"]]
                convs_no_answer = [
                    tokenize_with_prompt(conv, self.tokenizer, self.cfg.prompt_format) for conv in convs_no_answer
                ]
                answer_ids = self.generate(
                    prompts=left_collate_vectors(
                        [c.input_ids for c in convs_no_answer], padding_value=self.text_pad_id
                    ).to(self.device),
                    audios=dataset_batch["audios"].to(self.device, non_blocking=True),
                    audio_lens=dataset_batch["audio_lens"].to(self.device, non_blocking=True),
                    generation_config=GenerationConfig(
                        max_new_tokens=128,
                        bos_token_id=self.text_bos_id,
                        eos_token_id=[self.text_eos_id],
                        pad_token_id=self.text_pad_id,
                        do_sample=False,
                        num_beams=1,  # greedy decoding
                    ),
                )
                answer_ids = answer_ids.cpu()
                answer_ids = [parse_hyp(ans, [self.text_eos_id]) for ans in answer_ids]
                batch_answers = [self.tokenizer.ids_to_text(ans) for ans in answer_ids]
                for conv, ans in zip(convs_no_answer, batch_answers):
                    conv.turns.append(TextTurn(role="assistant", value=ans))
                    for k, v in list(conv.custom.items()):
                        if isinstance(v, torch.Tensor):
                            del conv.custom[k]
                    self._val_generations[name].append(conv.to_dict())

    def on_test_epoch_start(self) -> None:
        return self.on_validation_epoch_start()

    def on_test_epoch_end(self) -> None:
        return self.on_validation_epoch_end()

    def test_step(self, *args: Any, **kwargs: Any):
        return self.validation_step(*args, **kwargs)

    def backward(self, *args, **kwargs):
        with loss_parallel():
            super().backward(*args, **kwargs)

    @torch.no_grad()
    def generate(
        self,
        prompts: list[list[dict[str]]] | torch.Tensor,
        audios: torch.Tensor = None,
        audio_lens: torch.Tensor = None,
        generation_config: GenerationConfig = None,
        **generation_kwargs,
    ) -> torch.Tensor:
        """
        Generate LLM answers given text or mixed text+audio prompts.

        Example 1. High-level API using ``prompts`` to provide both text and audio::

            >>> answer_ids = model.generate(
            ...    prompts=[
            ...        [
            ...             {
            ...                 "role": "user",
            ...                 "content": f"Transcribe the following: {model.audio_locator_tag}",
            ...                 "audio": ["path/to/audio.wav"],
            ...             }
            ...         ]
            ...    ],
            ...    max_new_tokens=128,
            ... )

        You may also include a ``transformers.GenerationConfig`` object to customize decoding strategy::

            >>> answer_ids = model.generate(..., generation_config=GenerationConfig(do_sample=True, num_beams=5))

        Example 2. Lower-level API, using ``prompts`` for the text part,
        and pre-loaded ``audio`` and ``audio_lens`` tensors::

            >>> answer_ids = model.generate(
            ...    prompts=[
            ...        [{"role": "user", "content": f"Transcribe the following: {model.audio_locator_tag}"}],
            ...        [{"role": "user", "content": f"Transcribe the following in Polish: {model.audio_locator_tag}"}],
            ...    ],
            ...    audios=audios,  # torch.Tensor, float32, of shape (batch, time)
            ...    audio_lens=audio_lens,  # torch.Tensor, int64, of shape (batch,)
            ...    max_new_tokens=128,
            ... )

        Example 3. Lower-level API, using pre-tokenized and pre-formatted ``prompts`` for the text part,
        and pre-loaded ``audio`` and ``audio_lens`` tensors::

            >>> answer_ids = model.generate(
            ...    prompts=prompts,  # torch.Tensor, int64, of shape (batch, num_tokens)
            ...    audios=audios,  # torch.Tensor, float32, of shape (batch, time)
            ...    audio_lens=audio_lens,  # torch.Tensor, int64, of shape (batch,)
            ...    max_new_tokens=128,
            ... )

        Inputs:
            prompts: batch of prompts Tensor or as list[dict] each in the following format
                [
                  # batch example id 0
                  [{"role": "user"}, "slots": {"message": f"Transcribe the following: {model.audio_locator_tag}"}]
                  # batch example id 1
                  [{"role": "user"}, "slots": {"message": f"Transcribe the following in Polish: {model.audio_locator_tag}"}]
                ]
                "role" is LLM-specific, you can pass multiple turns as well.
                If ``prompts`` is a Tensor, we assume it was already formatted in the relevant chat template
                and tokenized with the model's tokenizer.
            audios: Optional. Time-domain audio signal zero-padded batch of shape (B, T).
                The number of audios must correspond to the number of occurrences of <audio_locator_tag> in prompts.
                Each prompt can have multiple audios.
            audio_lens: Optional. Length of each audio example.
            generation_config: Optional HuggingFace GenerationConfig object.
            generation_kwargs: Keyword arguments passed directly to the underlying LLM's ``generate`` method.
        """
        # Encode prompt dicts into int token ids.
        if isinstance(prompts, torch.Tensor):
            tokens = prompts
        else:
            if (
                maybe_audio := _resolve_audios_in_prompt(prompts, sampling_rate=self.sampling_rate, device=self.device)
            ) is not None:
                assert (
                    audios is None and audio_lens is None
                ), "Audios cannot be provided via ``prompts`` and ``audios``/``audio_lens`` arguments simultaneously."
                audios, audio_lens = maybe_audio
            formatter = PromptFormatter.resolve(self.cfg.prompt_format)(self.tokenizer)
            tokens = left_collate_vectors(
                [formatter.encode_dialog(turns=prompt)["input_ids"] for prompt in prompts],
                padding_value=self.text_pad_id,
            ).to(self.device)
        tokens_to_embed = tokens.where(tokens != self.audio_locator_tag_id, 0)
        token_embeds = self.embed_tokens(tokens_to_embed)

        if audios is not None:
            # Process audio when available
            # TODO: temporary workaround to perform batch_size=1 inference for audio encoder
            #   due to accuracy issues at bs>1
            # audio_embeds, audio_embed_lens = self.perception(audios, audio_lens)
            # audio_embeds = [audio_embeds[i, :elen] for i, elen in enumerate(audio_embed_lens)]

            encoded, encoded_len = self.perception.forward_encoder(input_signal=audios, input_signal_length=audio_lens)
            asr_hyps = self.perception.transcribe_encoded(encoded=encoded, encoded_len=encoded_len)
            asr_tokens = [
                torch.as_tensor(self.tokenizer.text_to_ids(f">> {hyp.text} <<" if hyp.text else ">> <<"))
                for hyp in asr_hyps
            ]
            asr_tokens_len = [at.shape[0] for at in asr_tokens]
            asr_tokens = torch.cat(asr_tokens, dim=0).unsqueeze(0).to(self.device)
            transcript_embs = torch.split(self.embed_tokens(asr_tokens).squeeze(0), asr_tokens_len, dim=0)
            audio_embeds, audio_embed_lens = self.perception(encoded=encoded, encoded_len=encoded_len)
            audio_embeds = [
                torch.cat([aemb[:aemblen], temb], dim=0)
                for aemb, aemblen, temb in zip(audio_embeds, audio_embed_lens, transcript_embs)
            ]
            # Insert audio embeddings into relevant positions in text embeddings.
            input_embeds, _, attention_mask = replace_placeholders_and_build_targets(
                input_ids=tokens,
                embeds=token_embeds,
                padding_id=self.text_pad_id,
                placeholder_id=self.audio_locator_tag_id,
                replacements=audio_embeds,
                target_ids=None,
            )
        else:
            # Text-only with embeddings - no audio placeholders to replace
            input_embeds = token_embeds
            attention_mask = tokens != self.text_pad_id

        generation_inputs = {"inputs_embeds": input_embeds, "attention_mask": attention_mask}
        if generation_config is None:
            generation_config = GenerationConfig(
                bos_token_id=self.text_bos_id,
                eos_token_id=self.text_eos_id,
                pad_token_id=self.text_pad_id,
            )
        # Generate the answers using HF Generate API.
        # Note: we need to put the text embedding layer back to the LLM for processing.
        with move_embedding(self):
            answer_tokens = self.llm.generate(
                **generation_inputs,
                **generation_kwargs,
                generation_config=generation_config,
            )
        return answer_tokens

    def configure_optimizers(self):
        return configure_optimizers(self)

    def configure_model(self) -> None:
        # TODO(pzelasko): refactor into separate module re-usable across models
        device_mesh = self.device_mesh
        if device_mesh is None:
            return

        llm = self.llm
        if isinstance(llm, PeftModel):
            llm = llm.base_model.model

        if (tp_mesh := device_mesh["tensor_parallel"]).size() > 1:
            self._use_tp = True

            # TODO: Distributing embeddings with TP in this setup is tricky
            #       because we're adding with the output of a non-parallelized
            #       speech encoder.
            # for m in (self.embed_tokens, self.embed_audio_tokens):
            #     parallelize_module(
            #         m,
            #         tp_mesh,
            #         ColwiseParallel(
            #             # input_layouts=Shard(1),
            #             # # Optional: Shard the output along the class dimension to compute the loss in parallel.
            #             # # See `loss_parallel` in `train.py`
            #             # output_layouts=Shard(1),
            #             # use_local_output=False,
            #         ),
            #     )

            # # Parallelize the first embedding and the last linear out projection
            plan = {
                "layers.0": PrepareModuleInput(
                    input_layouts=(Replicate(),),  # , None)
                    desired_input_layouts=(Shard(1),),  # , None)
                    use_local_output=True,
                ),
                "norm": SequenceParallel(),
            }
            parallelize_module(llm, tp_mesh, plan)

            # Parallelize each transformer block
            for transformer_block in llm.model.layers:
                plan = {
                    "input_layernorm": SequenceParallel(),
                    "self_attn.q_proj": ColwiseParallel(),
                    "self_attn.k_proj": ColwiseParallel(),
                    "self_attn.v_proj": ColwiseParallel(),
                    "self_attn.o_proj": RowwiseParallel(output_layouts=Shard(1)),
                    "post_attention_layernorm": SequenceParallel(),
                    "mlp": PrepareModuleInput(
                        input_layouts=(Shard(1),),
                        desired_input_layouts=(Replicate(),),
                    ),
                    "mlp.gate_proj": ColwiseParallel(),
                    "mlp.up_proj": ColwiseParallel(),
                    "mlp.down_proj": RowwiseParallel(output_layouts=Shard(1)),
                    # "pre_feedforward_layernorm": SequenceParallel(),
                    # "post_feedforward_layernorm": SequenceParallel(),
                }

                # Adjust attention module to use the local number of heads
                attn_layer = transformer_block.self_attn
                for attr in ("num_heads", "num_key_value_heads", "hidden_size"):
                    val = getattr(attn_layer, attr)
                    if val % tp_mesh.size() != 0:
                        logging.warning(
                            f"attn_layer.{attr}={val} is not divisible by {tp_mesh.size()=}: set a different tensor parallelism size to avoid errors."
                        )
                    setattr(attn_layer, attr, val // tp_mesh.size())

                # Apply the plan for the current transformer block
                parallelize_module(transformer_block, tp_mesh, plan)

            parallelize_module(
                llm.lm_head,
                tp_mesh,
                ColwiseParallel(
                    input_layouts=Shard(1),
                    # Optional: Shard the output along the class dimension to compute the loss in parallel.
                    # See `loss_parallel` in `train.py`
                    output_layouts=Shard(-1),
                    use_local_output=False,
                ),
            )

        if (dp_mesh := device_mesh["data_parallel"]).size() > 1:
            assert dp_mesh.ndim == 1  # Hybrid-sharding not supported
            self._use_fsdp = True
            fsdp_config = {"mesh": dp_mesh}
            for idx, layer in enumerate(llm.model.layers):
                llm.model.layers[idx] = fully_shard(layer, **fsdp_config)
            self.embed_tokens = fully_shard(self.embed_tokens, **fsdp_config)
            llm.lm_head = fully_shard(llm.lm_head, **fsdp_config)
            self.llm = fully_shard(self.llm, **fsdp_config)
            # self.perception.modality_adapter = fully_shard(self.perception.modality_adapter, **fsdp_config)
            # self.perception.asr.preprocessor = fully_shard(self.perception.asr.preprocessor **fsdp_config)
            # self.perception.asr.encoder = fully_shard(self.perception.asr.encoder, **fsdp_config)
            self.perception = fully_shard(self.perception, **fsdp_config)
            register_fsdp_forward_method(self.perception, "forward_encoder")
            register_fsdp_forward_method(self.perception, "transcribe_encoded")

    @property
    def oomptimizer_schema(self) -> dict:
        """
        Return a typing schema for optimal batch size calibration for various
        sequence lengths using OOMptimizer.
        """
        return {
            "cls": dict,
            "inputs": [
                {"name": "audios", "type": NeuralType(("B", "T"), AudioSignal()), "seq_length": "input"},
                {"name": "audio_lens", "type": NeuralType(("B",), LengthsType()), "seq_length": "input"},
                {
                    "name": "input_ids",
                    "type": NeuralType(("B", "T"), LabelsType()),
                    "seq_length": "output",
                    "vocab_size": self.text_vocab_size,
                },
                {"name": "loss_mask", "type": NeuralType(("B", "T"), MaskType()), "seq_length": "output"},
            ],
        }


def setup_speech_encoder_with_asr(model: torch.nn.Module, pretrained_weights: bool = True):
    """
    Sets up an ``AudioPerceptionModule``, initializing its ``encoder`` and ``preprocessor``
    with a pretrained NeMo ``ASRModel``.
    The result is assigned to ``model.perception`` attribute and is trainable.
    """
    with open_dict(model.cfg):
        model.cfg.output_dim = model.llm.config.hidden_size
    model.perception = AudioTranscriptionPerceptionModule(model.cfg.perception, model.cfg.pretrained_asr).train()

    from nemo.collections.common.parts.optional_cuda_graphs import WithOptionalCudaGraphs

    WithOptionalCudaGraphs.disable_cuda_graphs_recursive(model.perception.asr, attribute_path="decoding.decoding")


def parse_hyp(answer: torch.Tensor, eos_tokens: list[int]):
    end = torch.isin(answer, torch.tensor(eos_tokens)).nonzero(as_tuple=True)[0]
    if end.numel() == 0:
        return answer
    end = end[0]
    return answer[:end]


def strip_response_if_any(
    conversation: NeMoMultimodalConversation,
) -> NeMoMultimodalConversation:
    turns = conversation.turns
    while turns[-1].role == "assistant":
        turns = turns[:-1]
    return fastcopy(conversation, turns=turns)



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class MLFlowParams:
    """ML Flow Configuration Dataclass."""

    # name of experiment, if none, defaults to the globally set experiment name
    experiment_name: Optional[str] = None
    run_name: Optional[str] = None
    # if no run_name is set, it's set by version
    # local or remote tracking seerver. If tracking_uri is not set, it defaults to save_dir
    tracking_uri: Optional[str] = None
    tags: Optional[Dict[str, Any]] = None
    save_dir: Optional[str] = "./mlruns"
    prefix: str = ""
    artifact_location: Optional[str] = None
    # provide run_id if resuming a previously started run
    run_id: Optional[str] = None
    # Log checkpoints created by ModelCheckpoint as MLFlow artifacts.
    log_model: bool = False
